from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import binned_statistic
from scipy import interpolate
from scipy import signal
from scipy import fftpack
from submm.KIDs import calibrate
import pickle
from submm.KIDs import PCA_implementation as PCA
from tqdm import tqdm


def calibrate_multi(iq_sweep_data_f, iq_sweep_data_z, stream_f,stream_z,tau = 66*10**-9,
                    skip_beginning=0, plot_period=10, decimate=1, outfile_dir="./",
                    sample_rate=1.6*10**6, plot=True,n_comp_PCA = 0,verbose = True,
                    rotation_mode = 1,**keywords):
    '''
    function for calibrating data for an array of detectors i.e. turning 
    raw timestreams of IQ data in fractional frequency units df/f

    Parameters
    ----------
    iq_sweep_data_f: numpy array shape n_pts_iq x n_res
        The frequencies of an IQ calibration sweep
    iq_sweep_data_z: numpy array  shape n_pts_iq x n_res (complex with I = real Q = imag)
        Complex IQ swep data
    stream_f: list or numpy array length n_res
        For each streaming data series at what frequency was the tone
    stream_z: numpy array shape stream_length x n_res
        IQ data streamed
    decimate: int
        factor to bin the data by
    outfile_dir: str
        where to save the file
    sample_rate: float
        The frequency at which stream data was taken in Hz
    plot: bool
        toggle making a pdf plot of the calibration process
    n_comp_PCA: int
        Allow you to clean complex data by remove n components using 
        Priciple component analysis
    rotation_mode: int 1 or 2
        = 1 rotates streaming data to be at 0,1 in IQ plane 
        This maps frequendcy noise to real part and dissipation noise
        to the imaginary part (assumes noise ball is small)
        = 2 rotates data such that the furthest off res points
        in the IQ loop will be places at +/-pi. This makes interpolation 
        from phase to frequency not have a jump as phase goes from +pi to -pi

    Returns
    -------
    calibration dictionary with following keys
    
    fine_z: complex numpy array  shape n_pts_iq x n_re
        Copy of provided iq_sweep_data_z
    stream_z: numpy array shape stream_length x n_res
        Copy of provided stream_z
    iq_sweep_data_f: numpy array shape n_pts_iq x n_res  
        Copy of provided iq_sweep_data_f
    stream_corr:numpy array shape stream_length x n_res 
        stream z moved and rotated
    fine_corr: numpy array  shape n_pts_iq x n_res
        iq_sweep_data_z moved and rotated
    stream_df_over_f: numpy array shape stream_length x n_res 
        streaming data as df/f
    stream_time: numpy array shape stream_length
        time for each stream data point in seconds
    stream_corr_cleaned: numpy array shape stream_length x n_res or None
        cleaned IQ streaming data or None if data was not cleaned        
    stream_df_over_f_cleaned: numpy array shape stream_length x n_res 
        cleaned df/f data or None if data was not cleaned
    interp_functions: list of functions with len n_res
        functions used to turn phase into frequency during the calibration
    sample_rate: float 
        frequency of the streaming data in Hz
    '''

    stream_f = np.asarray(stream_f)
    
    #generate relative packet times
    stream_time = np.arange(0,stream_z.shape[0])[skip_beginning:]*1/sample_rate
    
    #bin the data if you like
    if decimate !=1:
        if verbose:
            print("decimating the data by factor of "+str(decimate))
        decimated_stream_z = stream_z #need deep copy?
        factors_of_10 = int(np.floor(np.log10(decimate)))
        for k in range(0,factors_of_10): # not suppose to decimate all at once
            decimated_stream_z = signal.decimate(decimated_stream_z,10,axis = 0)
            
        decimated_stream_z = signal.decimate(decimated_stream_z,
                                             decimate//(10**factors_of_10),axis =0)
        decimated_stream_time = np.arange(0,decimated_stream_z.shape[0])[skip_beginning:]*1/sample_rate*decimate

        stream_z = decimated_stream_z
        stream_time = decimated_stream_time
        sample_rate = sample_rate/decimate

    #initalize some arrays to hold the calibrated data
    stream_corr_all = np.zeros(stream_z.shape,dtype = 'complex')
    fine_corr_all = np.zeros(iq_sweep_data_z.shape,dtype = 'complex')
    stream_df_over_f_all = np.zeros(stream_z.shape)
    interp_functions = []
    circle_fit = np.ndarray((iq_sweep_data_z.shape[1],4))
    if n_comp_PCA>0:
        stream_df_over_f_all_cleaned = np.zeros(stream_z.shape)

    if verbose:
        print("Calibrating...")
    for k in tqdm(range(0,iq_sweep_data_z.shape[1]),ascii = True):
        
        #remove cable delay
        fine_corr = calibrate.remove_cable_delay(iq_sweep_data_f[:,k],iq_sweep_data_z[:,k],tau)
        stream_corr = calibrate.remove_cable_delay(stream_f[k],stream_z[:,k],tau)
        
        # fit a cicle to the data
        xc, yc, R, residu  = calibrate.leastsq_circle(np.real(fine_corr),np.imag(fine_corr))
        circle_fit[k,0:3] = np.array([xc, yc, R])

        #move the data to the origin
        fine_corr = fine_corr - xc -1j*yc
        stream_corr = stream_corr  - xc -1j*yc

        
        # rotate so streaming data is at 0 pi
        phase_stream = np.arctan2(np.imag(stream_corr),np.real(stream_corr))
        phase_fine = np.arctan2(np.imag(fine_corr),np.real(fine_corr))
        if rotation_mode == 2: # if you have data that covers a large part of the iq loop
            med_phase = np.arctan2(np.imag(fine_corr[0]+fine_corr[-1]),
                                    np.real(fine_corr[0]+fine_corr[-1])) + np.pi
            extrap = "min_max"
        elif rotation_mode == 1:
            med_phase = np.median(phase_stream)
            extrap = "extrapolate"
        else:
            print("pick valid rotation mode 1 or 2")
            return
        circle_fit[k,-1] = med_phase

        

        fine_corr_all[:,k] = fine_corr = fine_corr*np.exp(-1j*med_phase) 
        stream_corr_all[:,k] = stream_corr = stream_corr*np.exp(-1j*med_phase)
        
        stream_df_over_f_all[:,k],interp_function = \
            calibrate.interp_phase_to_df_over_f(phase_fine,
                                                phase_stream,
                                                iq_sweep_data_f[:,k],
                                                extrap = extrap)
        interp_functions.append(interp_function)

    if n_comp_PCA >0:
        interp_functions = []
        if verbose:
            print("cleaning data")
        cleaned, removed = PCA.PCA_SVD(stream_corr_all,n_comp_PCA,
                                       plot=True,sample_rate = sample_rate/decimate,
                                       outfile_dir = outfile_dir,plot_decimate = 100)
        stream_corr_all_cleaned = cleaned

        if verbose:
            print("computing cleaned df/f")
        for k in tqdm(range(0,iq_sweep_data_z.shape[1]),ascii = True):
            phase_fine = np.arctan2(np.imag(fine_corr_all[:,k]),np.real(fine_corr_all[:,k]))
            phase_stream = np.arctan2(np.imag(stream_corr_all_cleaned[:,k]),
                                      np.real(stream_corr_all_cleaned[:,k]))
            stream_df_over_f_all_cleaned[:,k],interp_function = \
                calibrate.interp_phase_to_df_over_f(phase_fine,
                                                    phase_stream,
                                                    iq_sweep_data_f[:,k],
                                                    extrap = extrap)
            interp_functions.append(interp_function)
    else:
        stream_corr_all_cleaned = None
        stream_df_over_f_all_cleaned = None
            
        
    #save everything to a dictionary
    cal_dict = {'fine_z': iq_sweep_data_z,
                'stream_z': stream_z,
                'iq_sweep_data_f':iq_sweep_data_f,
                'stream_corr':stream_corr_all,
                'fine_corr':fine_corr_all,
                'stream_df_over_f':stream_df_over_f_all,
                'stream_time':stream_time,
                'stream_corr_cleaned':stream_corr_all_cleaned,
                'stream_df_over_f_cleaned':stream_df_over_f_all_cleaned,
                'interp_functions':interp_functions,
                'sample_rate':sample_rate}

    #plot the data if desired
    if plot:
        if verbose:
            print("plotting")
        plot_calibrate(cal_dict, circle_fit,plot_period, outfile_dir)

    #save the dictionary
    if verbose:
        print("saving cal dict")
    pickle.dump( cal_dict, open(outfile_dir+ "cal.p", "wb" ) )
    return cal_dict


def plot_calibrate(cal_dict, circle_fit,plot_period, outfile_dir='./'):
    '''
    Function to plot the calibration process in general just called
    from calibrate_multitone

    Parameters                                                                                                                
    ----------
    cal_dict: calibration dictionary produced by calibrate_multitone
    circle_fit: numpy array with shape n_res x 4
        parameters that descibe a fitted circle
    plot_period: int
        plot only every nth point where n is plot period
    outfile_dir: str
        where to save the pdf

    Returns
    -------
    Nothing but produces a pdf plot called outfile_dir + cal_plots.pdf
    '''
    pdf_pages = PdfPages(outfile_dir+"cal_plots.pdf")
    for k in range(cal_dict['fine_z'].shape[1]):
        fig = plt.figure(k,figsize = (10,10))

        #plot the raw data
        plt.subplot(221,aspect = 'equal')
        plt.title("Raw data")
        plt.plot(np.real(cal_dict['fine_z'][:,k]),
                np.imag(cal_dict['fine_z'][:,k]),'o')
        plt.plot(np.real(cal_dict['stream_z'][:,k][::plot_period]),
                np.imag(cal_dict['stream_z'][:,k][::plot_period]),'.')
        
        
        #unpack corr data       
        fine_corr = cal_dict['fine_corr'][:,k]
        if cal_dict['stream_corr_cleaned'] is not None:
            stream_corr = cal_dict['stream_corr_cleaned'][:,k]
        else:
            stream_corr = cal_dict['stream_corr'][:,k]
        #reverse the circle offset and rotation
        med_phase = circle_fit[k,-1]
        xc = circle_fit[k,0]
        yc = circle_fit[k,1]
        R = circle_fit[k,2]
        fine_corr0 = fine_corr * np.exp(1.j*med_phase) + xc + 1.j * yc
        stream_corr0 = stream_corr * np.exp(1.j*med_phase) + xc + 1.j * yc
        

        plt.subplot(222)
        plt.title("Cable delay removed")
        plt.plot(np.real(fine_corr0),np.imag(fine_corr0),'o')
        plt.plot(np.real(stream_corr0)[10:-10][::plot_period],
                np.imag(stream_corr0)[10:-10][::plot_period],'.')

        #center and rotate IQ circle
        plt.subplot(223)
        plt.title("Moved to 0,0 and rotated")
        plt.plot(np.real(stream_corr)[2:-1][::plot_period],
                np.imag(stream_corr)[2:-1][::plot_period],'.')
        plt.plot(np.real(fine_corr),np.imag(fine_corr),'o')
        calibrate.plot_data_circle(np.real(fine_corr)-xc,np.imag(fine_corr)-yc,
                0, 0, R)

        #redo the phase fitting
        phase_fine = np.arctan2(np.imag(fine_corr),np.real(fine_corr))
        use_index = np.where((-np.pi/2.<phase_fine) & (phase_fine<np.pi/2))
        phase_stream = np.arctan2(np.imag(stream_corr),np.real(stream_corr))

        phase_small = np.linspace(np.min(phase_fine),np.max(phase_fine),1000)
        freqs_stream = cal_dict['interp_functions'][k](phase_stream)

        #plot the stream phase
        plt.subplot(224)
        plt.plot(phase_fine,cal_dict['iq_sweep_data_f'][:,k],'o--')
        plt.plot(phase_small,cal_dict['interp_functions'][k](phase_small),'--')
        plt.plot(phase_stream[::plot_period],freqs_stream[::plot_period],'.')
        ylim_upper = np.max(freqs_stream)+(np.max(freqs_stream)-np.min(freqs_stream))*3
        if ylim_upper>np.max(cal_dict['iq_sweep_data_f'][:,k]):
            ylim_upper = np.max(cal_dict['iq_sweep_data_f'][:,k])
        ylim_lower = np.min(freqs_stream)-(np.max(freqs_stream)-np.min(freqs_stream))*3
        if ylim_lower<np.min(cal_dict['iq_sweep_data_f'][:,k]):
            ylim_lower = np.min(cal_dict['iq_sweep_data_f'][:,k])
        plt.ylim(ylim_lower,ylim_upper)
        plt.xlim(np.min(phase_stream)-np.pi/4,np.max(phase_stream)+np.pi/4)
        plt.xlabel("phase")
        plt.ylabel("Frequency")

        pdf_pages.savefig(fig)
        plt.close(fig)

    pdf_pages.close()


def fft_noise(z_stream,df_over_f,sample_rate):
    '''
    Funtion for ffting resonator noise
    Parameters
    ----------
    z_stream: complex numpy array shape length streaming data x n_res
        IQ streaming data assume that the resonance has been moved
        to 0,0
    df_over_f: numpy array shape length streaming data x n_res
        streaming data in df/f units
    sample_rate: float
        sample rate at which streaming data was take in Hz
    
    Returns
    -------
    fft_freqs: numpy array
        The frequencies corresponding to the fft of Sxx,S_per,and S_par
    Sxx: numpy array
        The noise power spectral density in Sxx units where x
        is df/f
    S_per: numpy array
        the noise power spectral density perpendicular to the resonacne
        iq loop for KIDs with primarily response in frequency this will
        be a measure of detector noise in whatever units your readout 
        outputs per Hz i.e. volts^2/Hz
    S_par numpy array
        The noise power spectral density parrallel to the resonance
        circle. For KIDs with primarily response in frequency this will
        likely be a measure of the readout out noise. 
    '''
    npts_fft = int(2**(np.floor(np.log2(df_over_f.size)))) 
    Sxx = 2*fftpack.fft(df_over_f,n = npts_fft)*np.conj(fftpack.fft(df_over_f,n = npts_fft))/sample_rate*npts_fft/npts_fft**2
    #perpendicular should be radial on the circle
    per_stream = np.abs(z_stream)
    # radius times angle should be distance along the circle and should work if
    # noise ball is not small
    # so circumfernce*(phase/2pi) = 2*pi*radius*(phase/(2*pi)
    # then we also don't really have to rotate the noise ball to 1,0
    par_stream = np.mean(per_stream) * np.arctan2(np.imag(z_stream), np.real(z_stream))
    
    S_per = 2*fftpack.fft(per_stream,n = npts_fft)*np.conj(
            fftpack.fft(per_stream,n = npts_fft)
            )/sample_rate*npts_fft/npts_fft**2
    S_par = 2*fftpack.fft(par_stream,n = npts_fft)*np.conj(
            fftpack.fft(par_stream,n = npts_fft))/sample_rate*npts_fft/npts_fft**2
    fft_freqs = fftpack.fftfreq(npts_fft,1./sample_rate)
    return fft_freqs,Sxx,S_per,S_par


def noise_multi(cal_dict,outfile_dir = "./",n_comp_PCA = 0,plot = True,Sxx_ylims = None,verbose = True):
    '''
    Funciton to compute the noise power spectral density of resonators
    Parameters
    ----------
    cal_dict: dictionary
         calibration dictionary produced by calibrate_multitone 
    outfile_dir: str
        desired location of plots
    n_comp_PCA: int
        how many components to remove using pricipal component analysis on 
        the df/f data. If previously cleaned in cal this will be ignored
    plot: bool
        do or do not make plots of data
    Sxx_ylims: len 2 tuple or list
        when ploting min and max of y scale for Sxx
    verbose: bool
        print progress to terminal
    
    Returns
    -------
    Dictionary with the keys

    fft_freqs: numpy array
        The frequencies corresponding to the fft of Sxx,S_per,and S_par
    Sxx: numpy array
        The noise power spectral density in Sxx units where x
        is df/f
    S_per: numpy array
        the noise power spectral density perpendicular to the resonacne
        iq loop for KIDs with primarily response in frequency this will
        be a measure of detector noise in whatever units your readout
        outputs per Hz i.e. volts^2/Hz
    S_par: numpy array
        The noise power spectral density parrallel to the resonance
        circle. For KIDs with primarily response in frequency this will
        likely be a measure of the readout out noise.
    binned_freqs: numpy array
        binned version of fft_freqs
    Sxx_binned: numpy array
        binned version of Sxx
    S_per_binned: numpy array
        binned version of S_per
    S_par_binned: numpy array
        bineed version of S_par
    amp_subtracted: numpy array
        Sxx but renormalized using S_per and S_par to subtract of noise
        in the dissipation direction nominally attributed to the readout
    Sxx_clean: numpy array
        cleaned version of Sxx
    Sxx_binned_clean: numpy array
        binned and cleaned version of Sxx
    S_per_binned_clean: numpy array
        binned and cleaned version of S_per (requires cleaning in calibrate)
    S_par_binned_clean: numpy array
        binned and cleaned version of S_par (requires cleaning in calibrate)
    amp_subtracted_clean: numpy array
        binned and cleanead version of amp_subtracted
    '''
    sample_rate = cal_dict['sample_rate']
    if cal_dict['stream_df_over_f_cleaned'] is not None:
        do_PCA = True
        cleaned = cal_dict['stream_df_over_f_cleaned']
    elif n_comp_PCA >0:
        print("Consider doing PCA in calibration step")
        do_PCA = True
        #do PCA on the data
        cleaned, removed = PCA.PCA_SVD(cal_dict['stream_df_over_f'],n_comp_PCA,
                                       plot=True,sample_rate = sample_rate,outfile_dir = outfile_dir)
    else:
        do_PCA = False

    if verbose:
        print("FFTing...")
    for k in tqdm(range(0,cal_dict['fine_corr'].shape[1]),ascii=True):
        #print(k)
        
        fft_freqs,Sxx,S_per,S_par = fft_noise(cal_dict['stream_corr'][:,k],cal_dict['stream_df_over_f'][:,k],sample_rate)
        if do_PCA:
            if cal_dict['stream_corr_cleaned'] is not None:
                fft_freqs_2,Sxx_clean,S_per_clean,S_par_clean = fft_noise(cal_dict['stream_corr_cleaned'][:,k],cleaned[:,k], sample_rate)
            else:
                fft_freqs_2,Sxx_clean,S_per_clean,S_par_clean = fft_noise(cal_dict['stream_corr'][:,k],cleaned[:,k], sample_rate)
        if k == 0:
            #intialize some arrays
            Sxx_all = np.zeros((Sxx.shape[0],cal_dict['fine_corr'].shape[1]))
            Sxx_all_clean = np.zeros((Sxx.shape[0],cal_dict['fine_corr'].shape[1]))
            S_per_all = np.zeros((S_per.shape[0],cal_dict['fine_corr'].shape[1]))
            S_par_all = np.zeros((S_par.shape[0],cal_dict['fine_corr'].shape[1]))
            S_per_all_clean = np.zeros((S_per_clean.shape[0],cal_dict['fine_corr'].shape[1]))
            S_par_all_clean = np.zeros((S_par_clean.shape[0],cal_dict['fine_corr'].shape[1]))
            
        Sxx_all[:,k] = np.abs(Sxx)
        if do_PCA:
            Sxx_all_clean[:,k] = np.abs(Sxx_clean)
            S_per_all_clean[:,k] = np.abs(S_per_clean)
            S_par_all_clean[:,k] = np.abs(S_par_clean)
        S_per_all[:,k] = np.abs(S_per)
        S_par_all[:,k] = np.abs(S_par)

        # bin it for ploting
        #plot_bins = np.logspace(-3,np.log10(250),100)
        plot_bins = np.asarray(())
        smallest = fft_freqs[1]
        largest = np.max(fft_freqs)
        current = smallest
        while current < largest:
            plot_bins = np.append(plot_bins,10**np.floor(np.log10(current))*np.linspace(1.5,10.5,10))
            current = current*10
        binnedfreq =  binned_statistic(fft_freqs, fft_freqs, bins=plot_bins)[0] #bin the frequecy against itself
        binnedpsd = binned_statistic(fft_freqs, np.abs(Sxx), bins=plot_bins)[0]
        if do_PCA:
            binnedpsd_clean = binned_statistic(fft_freqs, np.abs(Sxx_clean), bins=plot_bins)[0]
            binnedper_clean = binned_statistic(fft_freqs, np.abs(S_per_clean), bins=plot_bins)[0]
            binnedpar_clean = binned_statistic(fft_freqs, np.abs(S_par_clean), bins=plot_bins)[0]
            amp_subtracted_clean = np.abs(binnedpsd_clean)*(binnedpar_clean-binnedper_clean)/binnedpar_clean
        binnedper = binned_statistic(fft_freqs, np.abs(S_per), bins=plot_bins)[0]
        binnedpar = binned_statistic(fft_freqs, np.abs(S_par), bins=plot_bins)[0]
        amp_subtracted = np.abs(binnedpsd)*(binnedpar-binnedper)/binnedpar
        if k == 0:
            Sxx_binned_all = np.zeros((binnedfreq.shape[0],cal_dict['fine_corr'].shape[1]))
            Sxx_binned_all_clean = np.zeros((binnedfreq.shape[0],cal_dict['fine_corr'].shape[1]))
            S_per_binned_all = np.zeros((binnedfreq.shape[0],cal_dict['fine_corr'].shape[1]))
            S_par_binned_all = np.zeros((binnedfreq.shape[0],cal_dict['fine_corr'].shape[1]))
            amp_subtracted_all = np.zeros((binnedfreq.shape[0],cal_dict['fine_corr'].shape[1]))
            S_per_binned_all_clean = np.zeros((binnedfreq.shape[0],cal_dict['fine_corr'].shape[1]))
            S_par_binned_all_clean = np.zeros((binnedfreq.shape[0],cal_dict['fine_corr'].shape[1]))
            amp_subtracted_all_clean = np.zeros((binnedfreq.shape[0],cal_dict['fine_corr'].shape[1]))

        Sxx_binned_all[:,k] = binnedpsd
        if do_PCA:
             Sxx_binned_all_clean[:,k] = binnedpsd_clean
             S_per_binned_all_clean[:,k] = binnedper_clean
             S_par_binned_all_clean[:,k] = binnedpar_clean
             amp_subtracted_all_clean[:,k] = amp_subtracted_clean
        S_per_binned_all[:,k] = binnedper
        S_par_binned_all[:,k] = binnedpar
        amp_subtracted_all[:,k] = amp_subtracted

    #make a psd dictionary
    psd_dict = {'fft_freqs':fft_freqs,
                    'Sxx':Sxx_all,
                    'S_per':S_per_all,
                    'S_par':S_par_all,
                    'binned_freqs':binnedfreq,
                    'Sxx_binned':Sxx_binned_all,
                    'S_per_binned':S_per_binned_all,
                    'S_par_binned':S_par_binned_all,
                    'amp_subtracted':amp_subtracted_all,
                    'Sxx_clean':Sxx_all_clean,
                    'Sxx_binned_clean':Sxx_binned_all_clean,
                    'S_per_binned_clean':S_per_binned_all_clean,
                    'S_par_binned_clean':S_par_binned_all_clean,
                    'amp_subtracted_clean':amp_subtracted_all_clean}

    #plot the stuff
    if verbose:
        print("plotting")
    if plot:
        plot_noise_multi(psd_dict, N_PCA=n_comp_PCA, outfile_dir=outfile_dir,Sxx_ylims = Sxx_ylims) 
    #save the psd dictionary
    if verbose:
        print("saving data")
    pickle.dump( psd_dict, open( outfile_dir+"psd.p", "wb" ) )

    return psd_dict


def plot_noise_multi(psd_dict, white_freq=100., N_PCA=0, outfile_dir = "./",Sxx_ylims = None):
    '''
    Funtion to plot noise in dictionary from noise multi
    
    Parameters
    ----------
    psd_dict: dictionary produced by noise_multi
    white_freq: float
        frequency at which to measure Sxx for combined plot os Sxx values
    N_PCA int:
        number of PCA components removed in cleaning for plot label
    outfile_dir: str
        where to put the pdf produced 
    Sxx_ylims: len 2 list or tuple
        ylima for Sxx plot

    Returns
    -------
    Nothing but produces a plot called outfile_dir +psd_plot.pdf
    '''
    #create the PDF file
    pdf_pages = PdfPages(outfile_dir+"psd_plots.pdf")
    #was there a PCA analysis
    do_PCA = (N_PCA > 0)
    #calculate white noise levels and plot them
    white_index = np.argmin(np.abs(psd_dict['fft_freqs']-white_freq))
    N_res = psd_dict['Sxx'].shape[1]
    Sxx_avg = np.ndarray((N_res,))
    par_per_ratio = np.ndarray((N_res,))
    for i in range(N_res):
        Sxx_avg[i] = psd_dict['Sxx'][white_index,i]
        par_per_ratio[i] = psd_dict['S_par'][white_index,i]/psd_dict['S_per'][white_index,i]
    fig = plt.figure(9000,figsize = (16,6))
    plt.subplot(122)
    plt.title("White Noise Levels")
    plt.semilogy(np.arange(N_res), Sxx_avg, 'bo')
    plt.ylabel("Sxx (1/Hz)")
    plt.xlabel("Resonator Index")
    plt.subplot(121)
    plt.title("Total to Amplifier Ratio")
    plt.plot(np.arange(N_res), par_per_ratio, 'go')
    plt.ylabel("Ratio of Parallel to Perpendicular")
    plt.xlabel("Resonator Index")
    pdf_pages.savefig(fig)
    plt.close(fig)
    #loop over the resonators
    for k in range(N_res):
        fig = plt.figure(k,figsize = (16,6))
        #plot Sxx raw and with amplifier subtracted
        plt.subplot(122)
        plt.title("Sxx")
        nan_index = np.isnan(psd_dict['binned_freqs'])
        if k == 0:
            plt.loglog(psd_dict['fft_freqs'],np.abs(psd_dict['Sxx'][:,k]))
        plt.loglog(psd_dict['binned_freqs'][~nan_index][0:-4],
                np.abs(psd_dict['Sxx_binned'][~nan_index][0:-4,k]),linewidth = 2,
                label = "Sxx raw")
        if do_PCA:
            plt.loglog(psd_dict['binned_freqs'][~nan_index][0:-4],
                    np.abs(psd_dict['Sxx_binned_clean'][~nan_index][0:-4,k]),linewidth = 2,
                    label = "PCA {0} comps".format(N_PCA))
        plt.loglog(psd_dict['binned_freqs'][~nan_index][0:-4],
                psd_dict['amp_subtracted'][~nan_index][0:-4,k],
                linewidth = 2, label="raw amp subtracted")
        plt.ylabel("Sxx (1/Hz)")
        plt.xlabel("Frequency (Hz)")
        plt.grid(which = "both")
        if Sxx_ylims is not None:
            plt.ylim(Sxx_ylims)
        plt.legend()

        #plot noise independant quadratures
        plt.subplot(121)
        plt.title("Res indx = "+str(k))

        plt.loglog(psd_dict['binned_freqs'][~nan_index][0:-4],psd_dict['S_per_binned'][~nan_index][0:-4,k],
                label = "amp noise")
        plt.loglog(psd_dict['binned_freqs'][~nan_index][0:-4],psd_dict['S_par_binned'][~nan_index][0:-4,k],
                label = "detect noise")
        plt.legend()
        plt.xlabel("Frequency (Hz)")
        plt.grid(which = "both")

        #save and close the figure
        pdf_pages.savefig(fig)
        plt.close(fig)

    #close the pdf file so everything gets flushed to the disk
    pdf_pages.close()



import numpy as np
from scipy.io.idl import readsav as readsav


# this is code to read in the mako multitone data sav files generated by Steve's idl code and store it in something useful for python


# Please note the date and details of any changes
# writen by Jordan Wheeler 12/13/2016
# add catch for instance when there are not

# To Do
# Would it be nicer to have these spit out a dictionary? or have another code that spits out everything as a python dictionary
# that would probably be better so that we don't have a backwards compatibility problem whenever we add new output.

# Change log
# 2017-06-02-jdw added catch for if there are no blind tone sin read_raw
# 2017-06-21-jdw changed file prefix so that it doesn't have to be in current directory
# 2017-06-21-jdw add keyword to allow loading of nonlinear fit extrapolated f0 calculated values

def read_raw(filename):  # filename is just the date/time code

    # Generate which resonators are blind and which are not so we don't look at blind tones
    # assumes it will be the same for each file
    raw = readsav(filename + "/rawdata.sav", verbose=False)
    bins = raw['multitone_data_raw']['bins'][0]
    blindbins = raw['multitone_data_raw']['blindbin'][0]
    if len(blindbins) == 0:  # there are no blind bins
        isblind = np.ones(bins.size)
    else:
        isblind = np.ones(bins.size)
        for i in range(0, bins.size):
            for j in range(0, blindbins.size):
                if bins[i] == blindbins[j]:
                    isblind[i] = 0

    blind_index = np.where(isblind < 0.5)[0]
    non_blind_index = np.where(isblind > 0)[0]

    f_res_all = raw['multitone_data_raw']['frlist'][0]
    # reads the data in 
    f_res = f_res_all[non_blind_index]  # only collect the f_res of the non blind bins
    f_res_blind = f_res_all[blind_index]  # only collect the f_res of the non blind bins

    raw_i = raw['multitone_data_raw'][0]['streamdata']['stream_data_concat'][0]['s21i'][0]
    raw_q = raw['multitone_data_raw'][0]['streamdata']['stream_data_concat'][0]['s21q'][0]
    print(raw_i.shape)
    non_blind_raw_i = raw_i[:, non_blind_index]
    non_blind_raw_q = raw_q[:, non_blind_index]

    blind_raw_i = raw_i[:, blind_index]
    blind_raw_q = raw_q[:, blind_index]

    # get the fine calibration data
    fine_i = np.zeros(
        (bins.shape[0], raw['multitone_data_raw']['fine_caldata'][0]['data_calibration'][0]['i'].shape[0]))
    fine_q = np.zeros(
        (bins.shape[0], raw['multitone_data_raw']['fine_caldata'][0]['data_calibration'][0]['q'].shape[0]))
    for b in range(0, raw['multitone_data_raw']['fine_caldata'][0]['data_calibration'][0]['i'].shape[0]):
        fine_i[:, b] = raw['multitone_data_raw']['fine_caldata'][0]['data_calibration'][0]['i'][b]
        fine_q[:, b] = raw['multitone_data_raw']['fine_caldata'][0]['data_calibration'][0]['q'][b]

    non_blind_fine_i = fine_i[non_blind_index, :]
    non_blind_fine_q = fine_q[non_blind_index, :]
    blind_fine_i = fine_i[blind_index, :]
    blind_fine_q = fine_q[blind_index, :]

    return non_blind_raw_i, non_blind_raw_q, blind_raw_i, blind_raw_q, f_res, f_res_blind, non_blind_fine_i, non_blind_fine_q, blind_fine_i, blind_fine_q


def read_cal(filename, **keywords):  # reading the calibrated file requires the raw file as well
    if ('model' in keywords):
        load_model = 1
    else:
        load_model = 0
    try:
        raw = readsav(filename + "/rawdata.sav", verbose=False)
    except:
        print("Reading the calibration file requires the raw file")
    cal = readsav(filename + "/calibrateddata.sav", verbose=False)

    # have to find out which bins are blind again
    bins = raw['multitone_data_raw']['bins'][0]
    blindbins = raw['multitone_data_raw']['blindbin'][0]
    isblind = np.ones(bins.size)

    for i in range(0, bins.size):
        for j in range(0, blindbins.size):
            if bins[i] == blindbins[j]:
                isblind[i] = 0

    blind_index = np.where(isblind < 0.5)[0]
    non_blind_index = np.where(isblind > 0)[0]

    S21_tan = np.imag(cal['multitone_data_calibrated'][0]['master_s21_stream_corr_rot_resampled']).astype('float64')[:,
              0, :]
    S21_norm = np.real(cal['multitone_data_calibrated'][0]['master_s21_stream_corr_rot_resampled']).astype('float64')[:,
               0, :]

    non_blind_S21_tan = S21_tan[:, non_blind_index]
    non_blind_S21_norm = S21_norm[:, non_blind_index]
    blind_S21_tan = S21_tan[:, blind_index]
    blind_S21_norm = S21_norm[:, blind_index]

    # f0_calc = cal['multitone_data_calibrated'][0]['master_f0_calculated_resampled'].astype('float64')[:,0,:]
    if load_model == 0:
        f0_calc = cal['multitone_data_calibrated'][0]['MASTER_F0_CALCULATED'].astype('float64')[:, 0, :]
    else:
        f0_calc = cal['multitone_data_calibrated'][0]['master_f0_calculated_fine_seriesx_nonlinear'].astype('float64')[
                  :, 0, :]
    try:
        nan_loc = np.where(np.isnan(f0_calc[:, -1]) == True)[0][0]
        f0_calc = f0_calc[0:nan_loc, :]
    except:
        f0_calc = cal['multitone_data_calibrated'][0]['master_f0_calculated_resampled'].astype('float64')[:, 0, :]

    f0_calc = f0_calc[:, non_blind_index]

    return non_blind_S21_tan, non_blind_S21_norm, blind_S21_tan, blind_S21_norm, f0_calc



import os
dir_this_file = os.path.dirname(os.path.realpath(__file__))
abs_path_parent_dir = dir_this_file.rsplit("submm", 1)[0]
abs_path_submm = os.path.join(abs_path_parent_dir, "submm")
abs_path_sample_data = os.path.join(abs_path_submm, "sample_data")
abs_path_output_dir_default = os.path.join(abs_path_submm, "output")


if __name__ == "__main__":
    print(f'abs_path_parent_dir:               {abs_path_parent_dir}')
    print(f'The absolute path for submm:       {abs_path_submm}')
    print(f'The absolute path for sample_data: {abs_path_sample_data}')
    print(f'The absolute path for output_dir:  {abs_path_output_dir_default}')



import os
import platform
from typing import NamedTuple

import numpy as np
import matplotlib.pyplot as plt
from scipy import signal, fftpack
from matplotlib.backends.backend_pdf import PdfPages
try:
    from PyQt5.QtCore import pyqtRemoveInputHook
    pyqtRemoveInputHook() #input() breaks during interactive plot on linux without this if using Qt backend
except:
    pass
from submm.KIDs.res.fitting import fit_nonlinear_iq, fit_nonlinear_mag

"""
Standalone version of kidPy's find_KIDs_interactive
use fin

If you have already identified most of the resonators at indexes kid_idx
just call the interactive plot object like so
ip = InteractivePlot(f,20*np.log10(np.abs(z)),kid_idx)

if you want to use use the filtering and threshold resonator finder
call find_vna_sweep(f,z) like
ip = find_vna_sweep(f,z)
get the kid indexes out from ip.kid_idx
get the frequencies out from f[ip.kid_idx]

"""


def open_stored_sweep(savepath, load_std=False):
    """Opens sweep data
       inputs:
           char savepath: The absolute path where sweep data is saved
       ouputs:
           numpy array Is: The I values
           numpy array Qs: The Q values"""
    files = sorted(os.listdir(savepath))
    I_list, Q_list, stdI_list, stdQ_list = [], [], [], []
    for filename in files:
        if filename.startswith('I'):
            I_list.append(os.path.join(savepath, filename))
        if filename.startswith('Q'):
            Q_list.append(os.path.join(savepath, filename))
        if filename.startswith('stdI'):
            stdI_list.append(os.path.join(savepath, filename))
        if filename.startswith('stdQ'):
            stdQ_list.append(os.path.join(savepath, filename))
    Is = np.array([np.load(filename) for filename in I_list])
    Qs = np.array([np.load(filename) for filename in Q_list])
    if len(stdI_list) > 0:
        std_Is = np.array([np.load(filename) for filename in stdI_list])
        std_Qs = np.array([np.load(filename) for filename in stdQ_list])
    if load_std:
        return Is, Qs, std_Is, std_Qs
    else:
        return Is, Qs


class SingleWindow(NamedTuple):
    left_max: int
    left_fitter_pad: int
    left_pad: int
    left_window: int
    minima: int
    right_window: int
    right_pad: int
    right_fitter_pad: int
    right_max: int


class InteractivePlot(object):
    """
    Convention is to supply the data in magnitude units i.e. 20*np.log10(np.abs(z))
    frequencies should be supplied in Hz
    """

    def __init__(self, chan_freqs, data, kid_idx, flags=None, f_old=None, data_old=None, kid_idx_old=None):
        plt.rcParams['keymap.forward'] = ['v']
        plt.rcParams['keymap.back'] = ['c', 'backspace']  # remove arrows from back and forward on plot
        plt.rcParams['keymap.quit'] = ['k']  # remove q for quit make it k for kill
        plt.rcParams['keymap.home'] = ['h']  # remove r for home only make it h
        plt.rcParams['keymap.fullscreen'] = ['shift+=']  # remove ('f', 'ctrl+f'), make +
        self.chan_freqs = chan_freqs
        self.data = data
        self.f_old = f_old
        self.data_old = data_old
        self.kid_idx_old = kid_idx_old
        print(kid_idx)
        self.kid_idx = kid_idx
        if flags is None:
            self.flags = []
            for i in range(len(self.kid_idx)): self.flags.append([])
        self.lim_shift_factor = 0.2
        self.zoom_factor = 0.1  # no greater than 0.5
        self.kid_idx_len = len(kid_idx)
        self.fig = plt.figure(1000, figsize=(16, 6))
        self.ax = self.fig.add_subplot(111)
        self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.fig.canvas.mpl_connect('key_release_event', self.on_key_release)
        self.fig.canvas.mpl_connect('button_press_event', self.onClick)
        self.l1, = self.ax.plot(self.chan_freqs / 10 ** 9, self.data)
        self.flagged_indexes = self.get_flagged_indexes()
        self.fp1, = self.ax.plot(self.chan_freqs[self.flagged_indexes] / 10 ** 9, self.data[self.flagged_indexes], "yo",
                                 markersize=9)
        self.p1, = self.ax.plot(self.chan_freqs[self.kid_idx] / 10 ** 9, self.data[self.kid_idx], "r*", markersize=8)
        self.text_dict = {}
        for i in range(0, len(self.kid_idx)):
            self.text_dict[i] = plt.text(self.chan_freqs[self.kid_idx][i] / 10 ** 9, self.data[self.kid_idx][i], str(i))

        if isinstance(self.f_old, np.ndarray):
            self.l2, = self.ax.plot(self.f_old / 10 ** 9, self.data_old, color="C0", alpha=0.25)
            self.p2, = self.ax.plot(self.f_old[self.kid_idx_old] / 10 ** 9, self.data_old[self.kid_idx_old], "r*",
                                    markersize=8,
                                    alpha=0.1)
            self.text_dict_old = {}
            for i in range(0, len(self.kid_idx_old)):
                self.text_dict_old[i] = plt.text(self.f_old[self.kid_idx_old][i] / 10 ** 9,
                                                 self.data_old[self.kid_idx_old][i],
                                                 str(i), color='Grey')

        self.shift_is_held = False
        self.control_is_held = False
        self.f_is_held = False
        self.add_list = []
        self.delete_list = []
        print("The controls are:")
        if platform.system() == 'Darwin':
            print("Hold the a key while right clicking to add points")
            print("Hold the d key while right clicking to delete points")
        else:
            print("Hold the shift key while right clicking to add points")
            print("Hold the control key while right clicking to delete points")
        print("Hold the f key while right clicking to flag points")
        print("Use the arrow keys to pan around plot")
        print("Use z to zoom")
        print("Use x to Xplode")
        print("Use q to zoom in x axis")
        print("Use w go Xplode in x axis")
        print("Use e to zoom in y axis")
        print("Use r go Xplode in y axis")
        print("Close all plots when finished")
        plt.xlabel('Frequency (GHz)')
        plt.ylabel('Power (dB)')
        plt.show(block=True)

    def on_key_press(self, event):
        # mac or windows
        if platform.system().lower() == 'darwin':
            if event.key == 'a':
                self.shift_is_held = True
            if event.key == 'd':
                self.control_is_held = True
        else:
            if event.key == 'shift':
                self.shift_is_held = True
            if event.key == 'control':
                self.control_is_held = True

        if event.key == 'f':
            self.f_is_held = True

        if event.key == 'right':  # pan right
            xlim_left, xlim_right = self.ax.get_xlim()
            xlim_size = xlim_right - xlim_left
            self.ax.set_xlim(xlim_left + self.lim_shift_factor * xlim_size,
                             xlim_right + self.lim_shift_factor * xlim_size)
            plt.draw()

        if event.key == 'left':  # pan left
            xlim_left, xlim_right = self.ax.get_xlim()
            xlim_size = xlim_right - xlim_left
            self.ax.set_xlim(xlim_left - self.lim_shift_factor * xlim_size,
                             xlim_right - self.lim_shift_factor * xlim_size)
            plt.draw()

        if event.key == 'up':  # pan up
            ylim_left, ylim_right = self.ax.get_ylim()
            ylim_size = ylim_right - ylim_left
            self.ax.set_ylim(ylim_left + self.lim_shift_factor * ylim_size,
                             ylim_right + self.lim_shift_factor * ylim_size)
            plt.draw()

        if event.key == 'down':  # pan down
            ylim_left, ylim_right = self.ax.get_ylim()
            ylim_size = ylim_right - ylim_left
            self.ax.set_ylim(ylim_left - self.lim_shift_factor * ylim_size,
                             ylim_right - self.lim_shift_factor * ylim_size)
            plt.draw()

        if event.key == 'z':  # zoom in
            xlim_left, xlim_right = self.ax.get_xlim()
            ylim_left, ylim_right = self.ax.get_ylim()
            xlim_size = xlim_right - xlim_left
            ylim_size = ylim_right - ylim_left
            self.ax.set_xlim(xlim_left + self.zoom_factor * xlim_size, xlim_right - self.zoom_factor * xlim_size)
            self.ax.set_ylim(ylim_left + self.zoom_factor * ylim_size, ylim_right - self.zoom_factor * ylim_size)
            plt.draw()

        if event.key == 'x':  # zoom out
            xlim_left, xlim_right = self.ax.get_xlim()
            ylim_left, ylim_right = self.ax.get_ylim()
            xlim_size = xlim_right - xlim_left
            ylim_size = ylim_right - ylim_left
            self.ax.set_xlim(xlim_left - self.zoom_factor * xlim_size, xlim_right + self.zoom_factor * xlim_size)
            self.ax.set_ylim(ylim_left - self.zoom_factor * ylim_size, ylim_right + self.zoom_factor * ylim_size)
            plt.draw()

        if event.key == 'q':  # zoom in x axis only
            xlim_left, xlim_right = self.ax.get_xlim()
            xlim_size = xlim_right - xlim_left
            self.ax.set_xlim(xlim_left + self.zoom_factor * xlim_size, xlim_right - self.zoom_factor * xlim_size)
            plt.draw()

        if event.key == 'w':  # zoom out x axis only
            xlim_left, xlim_right = self.ax.get_xlim()
            xlim_size = xlim_right - xlim_left
            self.ax.set_xlim(xlim_left - self.zoom_factor * xlim_size, xlim_right + self.zoom_factor * xlim_size)
            plt.draw()

        if event.key == 'e':  # zoom in y axis only
            ylim_left, ylim_right = self.ax.get_ylim()
            ylim_size = ylim_right - ylim_left
            self.ax.set_ylim(ylim_left + self.zoom_factor * ylim_size, ylim_right - self.zoom_factor * ylim_size)
            plt.draw()

        if event.key == 'r':  # zoom out y axis only
            ylim_left, ylim_right = self.ax.get_ylim()
            ylim_size = ylim_right - ylim_left
            self.ax.set_ylim(ylim_left - self.zoom_factor * ylim_size, ylim_right + self.zoom_factor * ylim_size)
            plt.draw()

    def on_key_release(self, event):
        # windows or mac
        if platform.system() == 'Darwin':
            if event.key == 'a':
                self.shift_is_held = False
            if event.key == 'd':
                self.control_is_held = False
        else:
            if event.key == 'shift':
                self.shift_is_held = False
            if event.key == 'control':
                self.control_is_held = False

        if event.key == 'f':
            self.f_is_held = False

    def onClick(self, event):
        if event.button == 3:
            if self.shift_is_held:  # add point
                print("adding point", event.xdata)
                self.kid_idx.append(np.argmin(np.abs(self.chan_freqs - event.xdata * 10 ** 9)))
                self.flags.append([])
                zipped_lists = zip(self.kid_idx, self.flags)
                sorted_pairs = sorted(zipped_lists)
                tuples = zip(*sorted_pairs)
                self.kid_idx, self.flags = [list(tuple) for tuple in tuples]
                # self.kid_idx.sort()
                self.refresh_plot()
            elif self.control_is_held:  # delete point
                print("removing point", event.xdata)
                delete_index = np.argmin(np.abs(self.chan_freqs[self.kid_idx] - event.xdata * 10 ** 9))
                self.kid_idx.pop(delete_index)
                self.flags.pop(delete_index)
                self.refresh_plot()
            elif self.f_is_held:  # flag resonator
                print("flagging point", event.xdata)
                flag_index = np.argmin(np.abs(self.chan_freqs[self.kid_idx] - event.xdata * 10 ** 9))
                print("current flags: ", self.flags[flag_index])
                flag = input("Enter flag string: ").lower()
                if flag == "c":
                    flag = "collision"
                elif flag == "s":
                    flag = "shallow"
                else:
                    pass
                self.flags[flag_index].append(flag)
                print("Flags are now: ", self.flags[flag_index])
                self.refresh_plot()
            else:
                print("please hold either the shift or control key while right clicking to add or remove points")

    def refresh_plot(self):
        self.flagged_indexes = self.get_flagged_indexes()
        self.fp1.set_data(self.chan_freqs[self.flagged_indexes] / 10 ** 9, self.data[self.flagged_indexes])
        self.p1.set_data(self.chan_freqs[self.kid_idx] / 10 ** 9, self.data[self.kid_idx])
        for i in range(0, self.kid_idx_len):
            self.text_dict[i].set_text("")  # clear all of the texts
        self.text_dict = {}
        for i in range(0, len(self.kid_idx)):
            self.text_dict[i] = plt.text(self.chan_freqs[self.kid_idx][i] / 10 ** 9, self.data[self.kid_idx][i], str(i))
        self.kid_idx_len = len(self.kid_idx)
        plt.draw()

    def get_flagged_indexes(self):
        flags_empty = True
        for flag_list in self.flags:
            if not flag_list:
                pass
            else:
                flags_empty = False
        if not flags_empty:
            zipped = zip(self.kid_idx, self.flags)
            pairs_with_flags = [pair for pair in zipped if len(pair[1]) > 0]
            tuples = zip(*pairs_with_flags)
            flagged_indexes, just_flags = [list(tuple) for tuple in tuples]
            return flagged_indexes
        else:
            return []


class InteractiveThresholdPlot(object):
    def __init__(self, f_Hz, s21_mag, peak_threshold_dB, spacing_threshold_Hz=1.0e5,
                 window_pad_factor=1.2, fitter_pad_factor=5.0, debug_mode=False):
        plt.rcParams['keymap.forward'] = ['v']
        plt.rcParams['keymap.back'] = ['c', 'backspace']  # remove arrows from back and forward on plot
        self.peak_threshold_dB = peak_threshold_dB
        self.spacing_threshold_Hz = spacing_threshold_Hz

        self.window_pad_factor = window_pad_factor
        self.fitter_pad_factor = fitter_pad_factor
        self.f_Hz = f_Hz
        self.spacing_threshold_Q = np.mean(self.f_Hz) / self.spacing_threshold_Hz
        self.f_GHz = f_Hz * 1.0e-9
        self.s21_mag = s21_mag

        self.regions = None
        self.ilo = None
        self.local_minima = None
        self.minima_as_windows = None
        self.calc_regions()

        if not debug_mode:
            self.fig = plt.figure(2, figsize=(16, 6))
            self.ax = self.fig.add_subplot(111)
            self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)
            self.l1, = self.ax.plot(self.f_GHz, self.s21_mag)

            self.p1, = self.ax.plot(self.f_GHz[self.ilo], self.s21_mag[self.ilo], "r*")
            self.p2, = self.ax.plot(self.f_GHz[self.local_minima], self.s21_mag[self.local_minima], "b*")
            print("Press up or down to change the threshold by 0.1 dB or press t to enter a custom threshold value.")
            print("Press left or right to change the spacing threshold or y to enter a custom threshold value.")
            print("Close all plots when finished")
            self.ax.set_xlabel(
                F"Frequency (GHz)\n Spacing threshold {'%.2E' % self.spacing_threshold_Hz} Hz or Q~{'%.2E' % self.spacing_threshold_Q}")
            plt.ylabel('Power (dB)')
            self.ax.set_title(
                F"Threshold: 3 adjacent points under {'%2.2f' % self.peak_threshold_dB} dB.\n Found {'%.0f' % len(self.local_minima)} minima")
            plt.show(block=True)

    def on_key_press(self, event):
        # print event.key
        # has to be shifted and ctrl because remote viewers only forward
        # certain key combinations
        # print event.key == 'd'
        if event.key == 'up':
            self.peak_threshold_dB = self.peak_threshold_dB + 0.1
            self.refresh_plot()
        if event.key == 'down':
            self.peak_threshold_dB = self.peak_threshold_dB - 0.1
            self.refresh_plot()
        if event.key == 'left':
            self.spacing_threshold_Hz = self.spacing_threshold_Hz / 1.4
            self.spacing_threshold_Q = np.mean(self.f_Hz) / self.spacing_threshold_Hz
            self.refresh_plot()
        if event.key == 'right':
            self.spacing_threshold_Hz = self.spacing_threshold_Hz * 1.4
            self.spacing_threshold_Q = np.mean(self.f_Hz) / self.spacing_threshold_Hz
            self.refresh_plot()
        if event.key == 't':
            self.peak_threshold_dB = np.float(input("What threshold would you like in dB? "))
            self.refresh_plot()
        if event.key == 'y':
            self.spacing_threshold_Hz = np.float(input("What Spacing threshold would you like in Hz? "))
            self.refresh_plot()

    def refresh_plot(self):
        self.calc_regions()
        self.p1.set_data(self.f_GHz[self.ilo], self.s21_mag[self.ilo])
        self.p2.set_data(self.f_GHz[self.local_minima], self.s21_mag[self.local_minima])
        self.ax.set_title(
            F"Threshold: 3 adjacent points under {'%2.2f' % self.peak_threshold_dB} dB.\n Found {'%.0f' % len(self.local_minima)} minima")
        self.ax.set_xlabel(
            F"Frequency (GHz)\n Spacing threshold {'%.2E' % self.spacing_threshold_Hz} Hz or Q~{'%.2E' % self.spacing_threshold_Q}")
        plt.draw()

    def calc_regions(self):
        bool_threshhold = self.s21_mag < -1.0 * self.peak_threshold_dB
        # self.ilo = np.where(self.s21_mag < -1.0 * self.peak_threshold_dB)[0]
        self.ilo = []
        self.regions = []
        self.local_minima = []
        is_in_theshhold_last = False
        sub_region = []
        for test_index, is_in_theshhold in list(enumerate(bool_threshhold)):
            if is_in_theshhold:
                self.ilo.append(test_index)
                sub_region.append(test_index)
            else:
                if is_in_theshhold_last:
                    # when the last point was in, but not this point it is time to finish the old region
                    self.regions.append(sub_region)
                    sub_region = []
            is_in_theshhold_last = is_in_theshhold
        else:
            if sub_region:
                self.regions.append(sub_region)

        window_calc_data = []
        # calculate the local minima in a simple brute force method
        for region in self.regions:
            minima_this_region = []
            minima_this_region_index = []
            found_this_region = False
            if len(region) > 2:
                for region_index in range(len(region) - 2):
                    middle_region_index = region_index + 1
                    middle_data_index = region[middle_region_index]
                    left = self.s21_mag[region[region_index]]
                    middle = self.s21_mag[middle_data_index]
                    right = self.s21_mag[region[region_index + 2]]
                    if middle < left and middle <= right:
                        found_this_region = True
                        self.local_minima.append(middle_data_index)
                        minima_this_region.append(middle_data_index)
                        minima_this_region_index.append(middle_region_index)
            if found_this_region:
                window_calc_data.append((region, minima_this_region_index, minima_this_region))

        # calculate the resonator windows
        self.minima_as_windows = []
        data_index_minima_left = None
        single_window = None
        right_window_not_found = False
        data_index_bound = 0
        for region, minima_this_region_index, minima_this_region in window_calc_data:
            # deal with spacing conflicts in the same region
            minima_this_region, minima_this_region_index = \
                self.resolve_spacing_conflicts(minima_this_region=minima_this_region,
                                               minima_this_region_index=minima_this_region_index)
            data_index_region_bound_left = region[0]
            data_index_region_bound_right = region[-1]
            # combine minima in the same region with a spacing conflict
            for region_index, data_index_minima in zip(minima_this_region_index, minima_this_region):
                # halfway to the next resonator
                if single_window is not None:
                    data_index_bound = int(np.round((data_index_minima_left + data_index_minima) / 2))
                    if right_window_not_found:
                        single_window["right_max"] = single_window["right_pad"] = \
                            single_window["right_fitter_pad"] = single_window["right_window"] = data_index_bound
                    else:
                        single_window["right_max"] = data_index_bound
                        test_right_pad = single_window["minima"] \
                                         + int(np.round((single_window["right_window"] - single_window["minima"]) \
                                                        * self.window_pad_factor))
                        if single_window["right_max"] < test_right_pad:
                            single_window["right_pad"] = single_window["right_max"]
                        else:
                            single_window["right_pad"] = test_right_pad
                        test_right_fitter_pad = single_window["minima"] \
                                                + int(np.round((single_window["right_window"] - single_window["minima"]) \
                                                               * self.fitter_pad_factor))
                        if single_window["right_max"] < test_right_fitter_pad:
                            single_window["right_fitter_pad"] = single_window["right_max"]
                        else:
                            single_window["right_fitter_pad"] = test_right_fitter_pad

                    self.minima_as_windows.append(SingleWindow(**single_window))
                # the window where resonator is located
                if region_index == minima_this_region_index[0]:
                    data_index_boundary_left = data_index_region_bound_left
                else:
                    data_index_boundary_left = data_index_bound
                if region_index == minima_this_region_index[-1]:
                    data_index_boundary_right = data_index_region_bound_right
                    right_window_not_found = False
                else:
                    right_window_not_found = True
                if right_window_not_found:
                    single_window = {"left_max": data_index_bound, "left_window": data_index_boundary_left,
                                     "minima": data_index_minima}
                else:
                    single_window = {"left_max": data_index_bound, "left_window": data_index_boundary_left,
                                     "minima": data_index_minima, "right_window": data_index_boundary_right}
                # window padding
                test_left_pad = single_window["minima"] \
                                - int(np.round((single_window["minima"] - single_window["left_window"])
                                               * self.window_pad_factor))
                if test_left_pad < single_window["left_max"]:
                    single_window["left_pad"] = single_window["left_max"]
                else:
                    single_window["left_pad"] = test_left_pad
                test_left_fitter_pad = single_window["minima"] \
                                       - int(np.round((single_window["minima"] - single_window["left_window"])
                                                      * self.fitter_pad_factor))
                if test_left_fitter_pad < single_window["left_max"]:
                    single_window["left_fitter_pad"] = single_window["left_max"]
                else:
                    single_window["left_fitter_pad"] = test_left_fitter_pad

                data_index_minima_left = single_window["minima"]
        else:
            # finish the last step in the loop
            data_index_bound = len(self.s21_mag)
            if right_window_not_found:
                single_window["right_max"] = single_window["right_window"] = data_index_bound
            else:
                single_window["right_max"] = data_index_bound
                test_right_pad = single_window["minima"] + \
                                 int(np.round((single_window["right_window"] - single_window["minima"])
                                              * self.window_pad_factor))
                if single_window["right_max"] < test_right_pad:
                    single_window["right_pad"] = single_window["right_max"]
                else:
                    single_window["right_pad"] = test_right_pad
                test_right_fitter_pad = single_window["minima"] \
                                        + int(np.round((single_window["right_window"] - single_window["minima"])
                                                       * self.fitter_pad_factor))
                if single_window["right_max"] < test_right_fitter_pad:
                    single_window["right_fitter_pad"] = single_window["right_max"]
                else:
                    single_window["right_fitter_pad"] = test_right_fitter_pad
            self.minima_as_windows.append(SingleWindow(**single_window))
        self.local_minima = [single_window.minima for single_window in self.minima_as_windows]
        # spacing conflicts across all regions
        self.local_minima, self.minima_as_windows = \
            self.resolve_spacing_conflicts(minima_this_region=self.local_minima,
                                           minima_this_region_index=self.minima_as_windows)

    def resolve_spacing_conflicts(self, minima_this_region, minima_this_region_index):
        found_spacing_conflict = True
        while found_spacing_conflict:
            found_spacing_conflict = False
            number_of_minima_this_region = len(minima_this_region)
            if number_of_minima_this_region > 1:
                for counter in range(number_of_minima_this_region - 1):
                    data_index_minima_left_test = minima_this_region[counter]
                    data_index_minima_right_test = minima_this_region[counter + 1]
                    minima_spacing_Hz = abs(
                        self.f_Hz[data_index_minima_left_test] - self.f_Hz[data_index_minima_right_test])
                    if minima_spacing_Hz < self.spacing_threshold_Hz:
                        # minima are too close:
                        print(F"Spacing Conflict in same threshold region.")
                        print(F"   Allowed spacing (MHz): {'%3.3f' % (self.spacing_threshold_Hz * 1.0e-6)}")
                        print(F"    Minima spacing (MHz): {'%3.3f' % (minima_spacing_Hz * 1.0e-6)}")
                        # keep the lowest of the minima
                        value_left_minima = self.s21_mag[data_index_minima_left_test]
                        value_right_minima = self.s21_mag[data_index_minima_right_test]
                        if value_left_minima < value_right_minima:
                            index_location_to_remove = counter + 1
                            index_location_to_keep = counter
                        else:
                            index_location_to_remove = counter
                            index_location_to_keep = counter + 1
                        # data for the print statement
                        data_index_kept = minima_this_region[index_location_to_keep]
                        data_index_removed = minima_this_region[index_location_to_remove]
                        value_kept_minima = self.s21_mag[data_index_kept]
                        f_MHz_kept_minima = self.f_GHz[data_index_kept] * 1.0e3
                        value_removed_minima = self.s21_mag[data_index_removed]
                        f_MHz_removed_minima = self.f_GHz[data_index_removed] * 1.0e3
                        # where the data is removed
                        minima_this_region_index.pop(index_location_to_remove)
                        minima_this_region.pop(index_location_to_remove)
                        # make the users see what decisions the code is making
                        print(F"Minima Kept: {value_kept_minima} dbM at {'%3.3f' % f_MHz_kept_minima} MHz")
                        print(F"Minima Removed: {value_removed_minima} dbM at {'%3.3f' % f_MHz_removed_minima} MHz\n")
                        # stop the loop here and restart from scratch with one less minima
                        found_spacing_conflict = True
                        break
        return minima_this_region, minima_this_region_index


def compute_dI_and_dQ(I, Q, freq=None, filterstr='SG', do_deriv=True):
    """
    Given I,Q,freq arrays
    input filterstr = 'SG' for sav-gol filter with builtin gradient, 'SGgrad' savgol then apply gradient to filtered
    do_deriv: if want to look at filtered non differentiated data.
    """
    if freq is None:
        df = 1.0
    else:
        df = freq[1] - freq[0]
    dI = filtered_differential(I, df, filtertype=filterstr, do_deriv=do_deriv)
    dQ = filtered_differential(Q, df, filtertype=filterstr, do_deriv=do_deriv)
    return dI, dQ


def filtered_differential(data, df, filtertype=None, do_deriv=True):
    """
    take 1d array data with spacing df. return filtered version of data depending on filterrype
    """
    window = 13
    n = 3
    if filtertype is None:
        out = np.gradient(data, df)
    elif filtertype.lower() == 'sg':
        if do_deriv == True:
            out = signal.savgol_filter(data, window, n, deriv=1, delta=df)
        else:
            out = signal.savgol_filter(data, window, n, deriv=0, delta=df)
    elif filtertype.lower() == 'sggrad':
        tobegrad = signal.savgol_filter(data, window, n)
        out = np.gradient(tobegrad, df)
    else:
        raise KeyError(F"filtertype: {filtertype} is not recognized.")
    return out


class InteractiveFilterPlot(object):
    def __init__(self, f_Hz, s21_mag, smoothing_scale_Hz=5.0e6):
        self.smoothing_scale_Hz = smoothing_scale_Hz
        self.f_Hz = f_Hz
        self.f_GHz = f_Hz * 1.0e-9
        self.s21_mag = s21_mag

        self.fig = plt.figure(2, figsize=(16, 6))
        self.ax = self.fig.add_subplot(111)
        self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)

        # first plot data and filter function before removing filter function
        self.filtermags = lowpass_cosine(y=self.s21_mag,
                                         tau=self.f_Hz[1] - self.f_Hz[0],
                                         f_3db=1. / self.smoothing_scale_Hz,
                                         width=0.1 * (1.0 / self.smoothing_scale_Hz),
                                         padd_data=True)
        # the cosine filter drops the last point is the array has an pdd number of points
        self.len_filtered = len(self.filtermags)
        self.s21_mag = self.s21_mag[:self.len_filtered]
        self.f_Hz = self.f_Hz[:self.len_filtered]
        self.f_GHz = self.f_Hz * 1.0e-9
        # calculations for peak spacing (rejection based on threshold)
        self.highpass_mags = self.s21_mag - self.filtermags
        self.l1, = self.ax.plot(self.f_GHz, self.s21_mag, label='#nofilter')
        self.l2, = self.ax.plot(self.f_GHz, self.filtermags, label='#filter')
        self.ax.set_title(F"Smoothing Scale: {'%.2E' % self.smoothing_scale_Hz} Hz.")
        plt.legend(loc=1)

        print(
            "Press left or right to change the smothing scale by 10% or press t to enter a custom smoothing scale value.")
        print("Close all plots when finished")
        plt.xlabel('Frequency (GHz)')
        plt.ylabel('Power (dB)')
        plt.show(block=True)

    def on_key_press(self, event):
        # print event.key
        # has to be shift and ctrl because remote viewers only forward
        # certain key combinations
        # print event.key == 'd'
        if event.key == 'left':
            self.smoothing_scale_Hz = self.smoothing_scale_Hz / 1.4
            self.refresh_plot()
        if event.key == 'right':
            self.smoothing_scale_Hz = self.smoothing_scale_Hz * 1.4
            self.refresh_plot()
        if event.key == 't':
            self.smoothing_scale_Hz = np.float(input("What smoothing scale would you like in Hz? "))
            self.refresh_plot()

    def refresh_plot(self):
        # first plot data and filter function before removing filter function
        self.filtermags = lowpass_cosine(y=self.s21_mag,
                                         tau=self.f_Hz[1] - self.f_Hz[0],
                                         f_3db=1. / self.smoothing_scale_Hz,
                                         width=0.1 * (1.0 / self.smoothing_scale_Hz),
                                         padd_data=True)
        # the cosine filter drops the last point is the array has an odd number of points
        self.len_filtered = len(self.filtermags)
        self.s21_mag = self.s21_mag[:self.len_filtered]
        self.f_Hz = self.f_Hz[:self.len_filtered]
        self.f_GHz = self.f_Hz * 1.0e-9
        # calculations for peak spacing (rejection based on threshold)
        self.highpass_mags = self.s21_mag - self.filtermags
        self.l1.set_data(self.f_GHz, self.s21_mag)
        self.l2.set_data(self.f_GHz, self.filtermags)
        self.ax.set_title(F"Smoothing Scale: {'%.2E' % self.smoothing_scale_Hz} Hz.")

        plt.draw()


def filter_trace(path, bb_freqs, sweep_freqs):
    chan_I, chan_Q = open_stored_sweep(path)
    channels = np.arange(np.shape(chan_I)[1])
    mag = np.zeros((len(bb_freqs), len(sweep_freqs)))
    chan_freqs = np.zeros((len(bb_freqs), len(sweep_freqs)))
    for chan in channels:
        mag[chan] = (np.sqrt(chan_I[:, chan] ** 2 + chan_Q[:, chan] ** 2))
        chan_freqs[chan] = (sweep_freqs + bb_freqs[chan]) / 1.0e6
    # mag = np.concatenate((mag[len(mag)/2:], mag[0:len(mag)/2]))
    mags = 20 * np.log10(mag / np.max(mag))
    mags = np.hstack(mags)
    # chan_freqs = np.concatenate((chan_freqs[len(chan_freqs)/2:],chan_freqs[0:len(chan_freqs)/2]))
    chan_freqs = np.hstack(chan_freqs)
    return chan_freqs, mags


def lowpass_cosine(y, tau, f_3db, width, padd_data=True):
    # padd_data = True means we are going to symmetric copies of the data to the start and stop
    # to reduce/eliminate the discontinuities at the start and stop of a dataset due to filtering
    #
    # False means we're going to have transients at the start and stop of the data
    # kill the last data point if y has an odd length
    if np.mod(len(y), 2):
        y = y[0:-1]
    # add the weird padd
    # so, make a backwards copy of the data, then the data, then another backwards copy of the data
    if padd_data:
        y = np.append(np.append(np.flipud(y), y), np.flipud(y))
    # take the FFT
    ffty = fftpack.fft(y)
    ffty = fftpack.fftshift(ffty)
    # make the companion frequency array
    delta = 1.0 / (len(y) * tau)
    nyquist = 1.0 / (2.0 * tau)
    freq = np.arange(-nyquist, nyquist, delta)
    # turn this into a positive frequency array
    # print((len(ffty) // 2))
    pos_freq = freq[(len(ffty) // 2):]
    # make the transfer function for the first half of the data
    i_f_3db = min(np.where(pos_freq >= f_3db)[0])
    f_min = f_3db - (width / 2.0)
    i_f_min = min(np.where(pos_freq >= f_min)[0])
    f_max = f_3db + (width / 2.0)
    i_f_max = min(np.where(pos_freq >= f_max)[0])
    transfer_function = np.zeros(len(y) // 2)
    transfer_function[0:i_f_min] = 1
    transfer_function[i_f_min:i_f_max] = (1 + np.sin(-np.pi * ((freq[i_f_min:i_f_max] - freq[i_f_3db]) / width))) / 2.0
    transfer_function[i_f_max:(len(freq) // 2)] = 0
    # symmetrize this to be [0 0 0 ... .8 .9 1 1 1 1 1 1 1 1 .9 .8 ... 0 0 0] to match the FFT
    transfer_function = np.append(np.flipud(transfer_function), transfer_function)
    # apply the filter, undo the fft shift, and invert the fft
    filtered = np.real(fftpack.ifft(fftpack.ifftshift(ffty * transfer_function)))
    # remove the padd, if we applied it
    if padd_data:
        filtered = filtered[(len(y) // 3):(2 * (len(y) // 3))]
    # return the filtered data
    return filtered


def find_vna_sweep(f_Hz, z, smoothing_scale_Hz=5.0e6, spacing_threshold_Hz=1.0e5):
    """
    f is frequencies (Hz)
    z is complex S21
    Smoothing scale (Hz)
    spacing threshold (Hz)
    """
    # first plot data and filter function before removing filter function
    s21_mag = 20 * np.log10(np.abs(z))
    ipf = InteractiveFilterPlot(f_Hz, s21_mag, smoothing_scale_Hz=smoothing_scale_Hz)

    # identify peaks using the interactive threshold plot
    ipt = InteractiveThresholdPlot(f_Hz=ipf.f_Hz,
                                   s21_mag=ipf.highpass_mags,
                                   peak_threshold_dB=1.5,
                                   spacing_threshold_Hz=spacing_threshold_Hz)

    # Zero everything but the resonators
    # highpass_mags[highpass_mags > -1.0 * ipt.peak_threshold_dB] = 0

    # the spacing thresholding was move to be inside the interactive threshold class
    kid_idx = ipt.local_minima
    ip = InteractivePlot(ipf.f_Hz, ipf.highpass_mags, kid_idx)
    return ip


def slice_vna(f, z, kid_index, q_slice=2000, flag_collided=True):
    # make f in Hz for fitting
    # Q = f/(delta f) for fitting is determined by the lowest frequencies assumed to be at index 0
    # delta f = f/Q
    df = f[1] - f[0]
    n_iq_points = int(f[0] / q_slice // df)
    if np.mod(n_iq_points, 2) == 0:
        n_iq_points = n_iq_points + 1
    print(n_iq_points)
    res_freq_array = np.zeros((n_iq_points, len(kid_index)))
    res_array = np.zeros((n_iq_points, len(kid_index))).astype('complex')
    print(res_array.dtype)
    for i in range(0, len(kid_index)):
        a = kid_index[i] - n_iq_points // 2 - 1
        b = kid_index[i] + n_iq_points // 2

        res_freq_array[:, i] = f[a:b]
        res_array[:, i] = z[a:b]
        if flag_collided:
            if i < len(kid_index) - 1:  # dont check last res
                # print(i)
                if kid_index[i + 1] - kid_index[i] < n_iq_points:  # collision at higher frequency
                    high_cutoff = int((kid_index[i + 1] + kid_index[i]) / 2)
                    # print(i,a,high_cutoff,b)
                    res_freq_array[high_cutoff - a:,i] = np.nan
                    res_array[high_cutoff - a:,i] = np.nan * (1 + 1j)
            if i != 0:  # dont check first res
                # print(i)
                if kid_index[i] - kid_index[i - 1] < n_iq_points:
                    low_cutoff = int((kid_index[i] + kid_index[i - 1]) / 2)
                    # print(i,a,low_cutoff,b)
                    res_freq_array[:low_cutoff - a, i] = np.nan
                    res_array[:low_cutoff - a, i] = np.nan * (1 + 1j)

    return res_freq_array, res_array


def fit_slices(res_freq_array, res_array, do_plots=True, plot_filename='fits'):
    if do_plots:
        pdf_pages = PdfPages(plot_filename + ".pdf")
    fits_dict_mag = {}
    fits_dict_iq = {}
    for i in range(0, res_freq_array.shape[0]):
        if do_plots:
            fig = plt.figure(i, figsize=(12, 6))
        try:
            fit = fit_nonlinear_iq(res_freq_array[i, :][~np.isnan(res_freq_array[i, :])],
                                   res_array[i, :][~np.isnan(res_array[i, :])])
            fits_dict_iq[i] = fit
            if do_plots:
                plt.subplot(121)
                plt.plot(np.real(res_array[i, :]), np.imag(res_array[i, :]), 'o', label='data')
                plt.plot(np.real(fit['fit_result']), np.imag(fit['fit_result']), label='fit')
                plt.plot(np.real(fit['x0_result']), np.imag(fit['x0_result']), label='guess')
                plt.legend()
        except Exception as inst:
            print("could not fit")
            print(type(inst))  # the exception instance
            print(inst.args)  # arguments stored in .args
            # print(inst)          # __str__ allows args to be printed directly,
            fits_dict_iq[i] = 'bad fit'
        try:
            fit2 = fit_nonlinear_mag(res_freq_array[i, :][~np.isnan(res_freq_array[i, :])],
                                        res_array[i, :][~np.isnan(res_array[i, :])])
            fits_dict_mag[i] = fit2
            if do_plots:
                plt.subplot(122)
                plt.plot(res_freq_array[i, :], 20 * np.log10(np.abs(res_array[i, :])), label='data')
                plt.plot(res_freq_array[i, :][~np.isnan(res_freq_array[i, :])],
                         10 * np.log10(np.abs(fit2['fit_result'])), label='fit')
                plt.plot(res_freq_array[i, :][~np.isnan(res_freq_array[i, :])],
                         10 * np.log10(np.abs(fit2['x0_result'])), label='guess')
                plt.legend()
        except Exception as inst:
            print("could not fit")
            print(type(inst))  # the exception instance
            print(inst.args)  # arguments stored in .args
            # print(inst)          # __str__ allows args to be printed directly,
            fits_dict_mag[i] = 'bad fit'
        if do_plots:
            pdf_pages.savefig(fig)
            plt.close()
    if do_plots:
        pdf_pages.close()

    return fits_dict_iq, fits_dict_mag


def retune_vna(f, z, kid_index, n_points_look_around=0, look_low_high=[0, 0], f_old=None, z_old=None,
               kid_index_old=None):
    """
    This is a program for when the resonances move and you need to retune the indexes of the resonators
    use n_point_look_around = 10 to look to lower and higher frequencies within 10 data points to find a new min
    use look_left_right = [10,20] to look for a new min 10 points to the lower frequencies and 20 points to higher frequencies

    if you would like to have the old data and kid indexes displayed in the background suppy
    f_old, z_old, kid_index old
    """
    if n_points_look_around > 0:
        for i in range(0, len(kid_index)):
            new_index = np.argmin(
                20 * np.log10(np.abs(z[kid_index[i] - n_points_look_around:kid_index[i] + n_points_look_around]))) + \
                        kid_index[i] - n_points_look_around
            kid_index[i] = new_index

    ip = InteractivePlot(f, 20 * np.log10(np.abs(z)), kid_index, f_old=f_old, data_old=20 * np.log10(np.abs(z_old)),
                         kid_idx_old=kid_index_old)

    return ip



import os
import sys
dir_this_file = os.path.dirname(os.path.realpath(__file__))
dir_notebook = os.path.join(dir_this_file, "res_fit.ipynb")
print("demo notebook located at")
print(dir_notebook)

try:
    import jupyterlab
except ModuleNotFoundError:
    input_str = input("This example requires a jupyterlab notebooks. Would you like to install that (y/n)?")
    if input_str.lower() == 'y':
        os.system("pip install jupyterlab")
    else:
        sys.exit("Exiting")


def run():
    os.system("jupyter notebook "+dir_notebook)




import pyvisa as visa

class temp_control(object):

    # Default constructor used to call the instrument at the given
    # GPIBaddress and initializes the object to the instrument
    def __init__(self):
        self.GPIBaddress = 'GPIB0::12::INSTR'
        rm = visa.ResourceManager("@py")
        self.obj1 = rm.open_resource(self.GPIBaddress)

    def get_temp(self,ch = 'A'):
        temp = float(self.obj1.query('INP '+  ch + ':TEMP?'))
        return temp


    def set_temp(self,temp,ch = 1):
        self.obj1.write('LOOP 1:SETP '+ str(temp))
        return

    def get_set_temp(self):
        set_temp = float(self.obj1.query('LOOP 1:SETP?')[:-1])
        return set_temp

    def turn_heater_on(self):
        self.obj1.write('CONTROL')
        return

    def turn_heater_off(self):
        self.obj1.write('STOP')
        return




import numpy as np
try:
    from astropy.io import fits
except:
    print("This function requires astropy to run") 

# this is code to read in the mako multitone data for fits file and store it in something useful for python


# Please not the date and details of any changes
# writen by Jordan Wheeler 8/25/2016
#Change log
# 3/17/2017 -Jordan added functionality to read stream fits files
# and added scaling the data by the scaling factor and averages



# read in calibration file i.e mako20160818_000147Cal.fits
def readcal(infile,hdu = False):
    #keyword blind
    # it would probably be good to just seperate the blind bins from the non blind bins
    # because I rarely actually care about the blind bins and it could simplfy my overarching codes
    
    hdulist = fits.open(infile)
    freqs = hdulist[1].data['freqs'][::1] # ther first data point is useless and dumb
    bins = hdulist[1].data['bins'][1::] #first data point is useless and dumb
    blind_bins = hdulist[1].data['Blindbin'][0:-1] #the last data point is useless and dumb
    I = hdulist[2].data['I']*hdulist[2].header['scaling']/hdulist[2].header['averages']
    Q = hdulist[2].data['Q']*hdulist[2].header['scaling']/hdulist[2].header['averages']
    I_Q_freqs = hdulist[2].header['bininHz']*hdulist[2].data['tones']

    # we want to know which bins are blind bins so that we can appropriatly ignore them
    # nested for loop is bad (slow in python) see if we can do better when bored
    isblind = np.ones(bins.shape[0])
    for k in range(0,blind_bins.size):
        for j in range(0,blind_bins.size):
            if bins[k] == blind_bins[j]:
                isblind[k] = -1

    #lets seperate the blind bins from the real bins
    blind_index = np.where(isblind<0)[0]
    non_blind_index = np.where(isblind>0)[0]

    I_blind = I[:,blind_index]
    Q_blind = Q[:,blind_index]
    freqs_blind = freqs[blind_index]
    I_Q_freqs_blind = I_Q_freqs[:,blind_index]

    I_res = I[:,non_blind_index]
    Q_res = Q[:,non_blind_index]
    freqs_res = freqs[non_blind_index]
    I_Q_freqs_res = I_Q_freqs[:,non_blind_index]

    #make a dicitonary for it
    cal_dict = {'I': I_res, 'Q': Q_res, 'freqs': I_Q_freqs_res, 'center_freqs': freqs_res,'I_blind':I_blind,'Q_blind':Q_blind,'freqs_blind':I_Q_freqs_blind,'center_freqs_blind':freqs_blind}
    
    hdulist.close()
    
    # we want to reformat some of the data so that it is more accessable
    if hdu == False:
        return cal_dict
    else:
         return hdulist  


def readstream(infile,hdu = False):

    hdulist = fits.open(infile)
    
    cal_file = hdulist[0].header['calfile']
    hdulist_cal = fits.open(cal_file)
    scaling = hdulist_cal[2].header['scaling']
    averages = hdulist_cal[2].header['averages']
    freqs = hdulist_cal[1].data['freqs'][::1] # ther first data point is useless and dumb
    bins = hdulist_cal[1].data['bins'][1::] #first data point is useless and dumb
    blind_bins = hdulist_cal[1].data['Blindbin'][0:-1] #the last data point is useless and dumb
    hdulist_cal.close()

    # we want to know which bins are blind bins so that we can appropriatly ignore them
    # nested for loop is bad (slow in python) see if we can do better when bored
    isblind = np.ones(bins.shape[0])
    for k in range(0,blind_bins.size):
        for j in range(0,blind_bins.size):
            if bins[k] == blind_bins[j]:
                isblind[k] = -1

    blind_index = np.where(isblind<0)[0]
    non_blind_index = np.where(isblind>0)[0]

    #we have to find out how many hdulist entrys there are
    i = -1
    test = False
    while test == False:
        try:
            i = i+1
            k = hdulist[i]
        except:
            test = True
    hdulist_length = i

    tones = hdulist[1].data['tones'][0,:]
    
    I = hdulist[1].data['I']*scaling/averages
    Q = hdulist[1].data['Q']*scaling/averages
    for i in range(2,hdulist_length):
        I = np.vstack((I,hdulist[i].data['I']*scaling/averages))
        Q = np.vstack((Q,hdulist[i].data['Q']*scaling/averages))

    I_blind = I[:,blind_index]
    Q_blind = Q[:,blind_index]

    I_res = I[:,non_blind_index]
    Q_res = Q[:,non_blind_index]

    stream_dict = {'I': I_res, 'Q': Q_res, 'I_blind':I_blind,'Q_blind':Q_blind}

    hdulist.close()

    # we want to reformat some of the data so that it is more accessable
    if hdu == False:
        return stream_dict
    else:
         return hdulist

def readsearch(filename,hdu = False):
    data  = fits.open(filename)
    #some stuff to figure howmany hdulist there are
    test = 0
    i = 0
    while test == 0:
        try:
            data[i].header[0]
            i = i +1
        except IndexError:
            test = 1

    num_rows = i 
    sweeps = (num_rows - 1)/2
    freqs = []
    S21 = []
    I = []
    Q = []


    for j in range(0,sweeps):# This is triple nested for loop So it is super slow becuase python sucks at nested for loops should fix with numba
        for i in range(0,data[1+j*2].data.shape[0]-1):
            for k in range(0,data[2+j*2].header['naxis2']):
                freqs = np.append(freqs,data[2+j*2].data['tones'][k,i]*data[2+j*2].header['bininHz'])
                I = np.append(I,data[2+j*2].data['I'][k,i]*data[2+j*2].header['scaling'])
                Q = np.append(Q,data[2+j*2].data['Q'][k,i]*data[2+j*2].header['scaling'])
                #S21 = np.append(S21,np.mean(data[2+j*2].data['I'][k,i]**2+data[2+j*2].data['Q'][k,i]**2)*data[2+j*2].header['scaling']**2)#Have to multiply by weird scaling factor
    #print I[0]
    sorted_data = np.vstack((freqs,I,Q))
    sorted_data = np.transpose(sorted_data)
    sorted_data = sorted_data[sorted_data[:,0].argsort()]

    freqs = sorted_data[:,0]
    I = sorted_data[:,1]
    Q = sorted_data[:,2]


    search_dict = {'freqs':freqs,'I': I, 'Q': Q}

    # we want to reformat some of the data so that it is more accessable
    if hdu == False:
        return search_dict
    else:
         return data




from matplotlib.backends.backend_pdf import PdfPages
from submm.multitone_kidPy import read_multitone
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import binned_statistic
from scipy import interpolate, fftpack
from submm.KIDs import calibrate, PCA_implementation as PCA
from submm.KIDs.res.fitting import fit_nonlinear_mag_sep, fit_nonlinear_iq_sep
from submm.KIDs.res.utils import amplitude_normalization_sep, guess_x0_mag_nonlinear_sep, guess_x0_iq_nonlinear_sep
import pickle


# this function fits a fine and gain scan combo produced by the ASU multi-tone system
# and uses the error produced by the system to determine if the fit is good
def fit_fine_gain_std(fine_name, gain_name, reduced_chi_squared_cutoff=1000., plot=True):
    fine = read_multitone.read_iq_sweep(fine_name, load_std=True)
    gain = read_multitone.read_iq_sweep(gain_name, load_std=True)
    outfile_dir = fine_name
    center_freqs = fine['freqs'][fine['freqs'].shape[0] // 2, :]

    if plot:
        pdf_pages = PdfPages(outfile_dir + "/" + "fit_plots.pdf")

    all_fits_mag = np.zeros((9, fine['freqs'].shape[1]))
    all_fits_iq = np.zeros((10, fine['freqs'].shape[1]))

    for i in range(0, fine['freqs'].shape[1]):
        fine_f = fine['freqs'][:, i] * 10 ** 6
        gain_f = gain['freqs'][:, i] * 10 ** 6
        fine_z = fine['I'][:, i] + 1.j * fine['Q'][:, i]
        fine_z_err = fine['I_std'][:, i] + 1.j * fine['Q_std'][:, i]
        gain_z = gain['I'][:, i] + 1.j * gain['Q'][:, i]
        gain_z_err = gain['I_std'][:, i] + 1.j * gain['Q_std'][:, i]

        # flag data that is too close to other resonators
        distance = center_freqs - center_freqs[i]
        if center_freqs[i] != np.min(center_freqs):  # don't do if lowest frequency resonator
            closest_lower_dist = -np.min(np.abs(distance[np.where(distance < 0)]))
            closest_lower_index = np.where(distance == closest_lower_dist)[0][0]
            halfway_low = (center_freqs[i] + center_freqs[closest_lower_index]) / 2.
        else:
            halfway_low = 0

        if center_freqs[i] != np.max(center_freqs):  # don't do if highest frequenct
            closest_higher_dist = np.min(np.abs(distance[np.where(distance > 0)]))
            closest_higher_index = np.where(distance == closest_higher_dist)[0][0]
            halfway_high = (center_freqs[i] + center_freqs[closest_higher_index]) / 2.
        else:
            halfway_high = np.inf

        use_index = np.where(((fine_f / 10 ** 6 > halfway_low) & (fine_f / 10 ** 6 < halfway_high)))
        fine_f = fine_f[use_index]
        fine_z = fine_z[use_index]
        fine_z_err = fine_z_err[use_index]

        # flag gain data that is to close to all of the resonators
        fine_span = (fine_f[-1] - fine_f[0]) / 10 ** 6
        use_boolean = np.zeros(len(gain_f))
        for j in range(0, len(gain_f)):
            if (np.abs(gain_f[j] / 10 ** 6 - center_freqs) < fine_span / 2.).any():
                use_boolean[j] = 1

        use_index_gain = np.where(use_boolean != 1)
        if len(use_index_gain[0] > 1):
            gain_f = gain_f[use_index_gain]
            gain_z = gain_z[use_index_gain]
            gain_z_err = gain_z_err[use_index_gain]

        if plot:
            fig = plt.figure(i, figsize=(16, 10))

            ax1 = plt.subplot(231)
            plt.title("Resonator Index " + str(i))
            plt.plot(fine['freqs'][:, i], 10 * np.log10(fine['I'][:, i] ** 2 + fine['Q'][:, i] ** 2), 'o', label="fine")
            plt.plot(gain['freqs'][:, i], 10 * np.log10(gain['I'][:, i] ** 2 + gain['Q'][:, i] ** 2), 'o', label="gain")
            plt.xlabel("Frequency (MHz)")
            plt.ylabel("Power (dB)")

            ax2 = plt.subplot(232)
            plt.plot(fine['freqs'][:, i], 10 * np.log10(fine['I'][:, i] ** 2 + fine['Q'][:, i] ** 2), 'o')
            plt.plot(gain['freqs'][:, i], 10 * np.log10(gain['I'][:, i] ** 2 + gain['Q'][:, i] ** 2), 'o')
            plt.xlabel("Frequency (MHz)")
            plt.ylabel("Power (dB)")
            plt.xlim(np.min(fine['freqs'][:, i]), np.max(fine['freqs'][:, i]))

        # fit nonlinear magnitude
        try:
            x0 = guess_x0_mag_nonlinear_sep(fine_f, fine_z, gain_f, gain_z, verbose=True)
            fit_dict_mag = fit_nonlinear_mag_sep(fine_f, fine_z, gain_f, gain_z, fine_z_err=fine_z_err,
                                                         gain_z_err=gain_z_err, x0=x0)  # ,bounds =bounds)
            all_fits_mag[0:8, i] = fit_dict_mag['fit'][0]
            all_fits_mag[8, i] = fit_dict_mag['red_chi_sqr']
            if plot:
                plt.subplot(231)
                plt.plot(fit_dict_mag['fit_freqs'] / 10 ** 6, 10 * np.log10(fit_dict_mag['fit_result']), "+",
                         label="fit")
                plt.plot(fit_dict_mag['fit_freqs'] / 10 ** 6, 10 * np.log10(fit_dict_mag['x0_result']), "x",
                         label="x0 guess")
                plt.title("f =" + str(fit_dict_mag['fit'][0][0] / 10 ** 6)[0:7] + "MHz, a=" + "{:.2f}".format(
                    fit_dict_mag['fit'][0][4]))
                plt.legend()
                if fit_dict_mag['red_chi_sqr'] > reduced_chi_squared_cutoff:
                    ax1.set_facecolor('lightyellow')
                    ax2.set_facecolor('lightyellow')

                plt.subplot(232)
                plt.plot(fit_dict_mag['fit_freqs'] / 10 ** 6, 10 * np.log10(fit_dict_mag['fit_result']), "+")
                plt.plot(fit_dict_mag['fit_freqs'] / 10 ** 6, 10 * np.log10(fit_dict_mag['x0_result']), "x")

                plt.text(0.75, 0.9, "fr  = " + "{:.3f}".format(fit_dict_mag['fit'][0][0] / 10 ** 6) + " MHz",
                         fontsize=14, transform=plt.gcf().transFigure)
                plt.text(0.75, 0.85, "Qr  = " + "{:.0f}".format(fit_dict_mag['fit'][0][1]) + " ", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.8, "amp = " + "{:.2f}".format(fit_dict_mag['fit'][0][2]) + " ", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.75, "phi = " + "{:.2f}".format(fit_dict_mag['fit'][0][3]) + " radians", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.70, "a   = " + "{:.2f}".format(fit_dict_mag['fit'][0][4]) + " ", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.65, "b0  = " + "{:.0f}".format(fit_dict_mag['fit'][0][5]) + " ", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.6, "b1  = " + "{:.0f}".format(fit_dict_mag['fit'][0][6]) + " ", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.55, "reduced chi squared  = " + "{:.2f}".format(fit_dict_mag['red_chi_sqr']) + " ",
                         fontsize=14, transform=plt.gcf().transFigure)

        except Exception as e:
            print(e)
            print("could not fit the resonator")

        if plot:
            ax3 = plt.subplot(234, aspect='equal')
            plt.plot(fine['I'][:, i], fine['Q'][:, i], 'o')
            plt.plot(gain['I'][:, i], gain['Q'][:, i], 'o')
            plt.xlabel("I")
            plt.ylabel("Q")

            ax4 = plt.subplot(235, aspect='equal')
            plt.plot(fine['I'][:, i], fine['Q'][:, i], 'o')
            plt.plot(gain['I'][:, i], gain['Q'][:, i], 'o')
            plt.xlabel("I")
            plt.ylabel("Q")
            plt.xlim(np.min(fine['I'][:, i]), np.max(fine['I'][:, i]))
            plt.ylim(np.min(fine['Q'][:, i]), np.max(fine['Q'][:, i]))

        # fit nonlinear iq
        try:
            x0 = guess_x0_iq_nonlinear_sep(fine_f, fine_z, gain_f, gain_z, verbose=True)
            fit_dict_iq = fit_nonlinear_iq_sep(fine_f, fine_z, gain_f, gain_z, fine_z_err=fine_z_err,
                                                       gain_z_err=gain_z_err, x0=x0)
            all_fits_iq[0:9, i] = fit_dict_iq['fit'][0]
            all_fits_iq[9, i] = fit_dict_iq['red_chi_sqr']
            if plot:
                plt.subplot(234, aspect='equal')
                plt.plot(np.real(fit_dict_iq['fit_result']), np.imag(fit_dict_iq['fit_result']), "+")
                plt.plot(np.real(fit_dict_iq['x0_result']), np.imag(fit_dict_iq['x0_result']), "x")
                plt.title("f =" + str(fit_dict_iq['fit'][0][0] / 10 ** 6)[0:7] + "MHz, a=" + "{:.2f}".format(
                    fit_dict_iq['fit'][0][4]))
                plt.subplot(235, aspect='equal')
                plt.plot(np.real(fit_dict_iq['fit_result']), np.imag(fit_dict_iq['fit_result']), "+")
                plt.plot(np.real(fit_dict_iq['x0_result']), np.imag(fit_dict_iq['x0_result']), "x")
                plt.plot(fine['I'][:, i], fine['Q'][:, i])

                if fit_dict_iq['red_chi_sqr'] > reduced_chi_squared_cutoff:
                    ax3.set_facecolor('lightyellow')
                    ax4.set_facecolor('lightyellow')

                plt.text(0.75, 0.45, "fr  = " + "{:.3f}".format(fit_dict_iq['fit'][0][0] / 10 ** 6) + " MHz",
                         fontsize=14, transform=plt.gcf().transFigure)
                plt.text(0.75, 0.4, "Qr  = " + "{:.0f}".format(fit_dict_iq['fit'][0][1]) + " ", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.35, "amp = " + "{:.2f}".format(fit_dict_iq['fit'][0][2]) + " ", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.3, "phi = " + "{:.2f}".format(fit_dict_iq['fit'][0][3]) + " radians", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.25, "a   = " + "{:.2f}".format(fit_dict_iq['fit'][0][4]) + " ", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.2, "i0  = " + "{:.0f}".format(fit_dict_iq['fit'][0][5]) + " ", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.15, "q0  = " + "{:.0f}".format(fit_dict_iq['fit'][0][6]) + " ", fontsize=14,
                         transform=plt.gcf().transFigure)
                plt.text(0.75, 0.1, "tau = " + "{:.2f}".format(fit_dict_iq['fit'][0][7] * 10 ** 7) + " x 10^-7 ",
                         fontsize=14, transform=plt.gcf().transFigure)
                plt.text(0.75, 0.05, "reduced chi squared  = " + "{:.2f}".format(fit_dict_iq['red_chi_sqr']) + " ",
                         fontsize=14, transform=plt.gcf().transFigure)

        except Exception as e:
            print(e)
            print("could not fit the resonator")

        if plot:
            plt.suptitle("Resonator index = " + str(i) + ", Frequency = " + str(center_freqs[i])[0:7])

            pdf_pages.savefig(fig)
            plt.close(fig)
    if plot:
        pdf_pages.close()

    pdf_pages = PdfPages(outfile_dir + "/" + "fit_results.pdf")

    fig = plt.figure(1, figsize=(12, 6))
    plt.title("Center frequency")
    res_index = np.arange(0, all_fits_mag.shape[1])

    plt.plot(res_index, all_fits_mag[0, :] / 10 ** 6, 'o', label="Mag fit")
    plt.plot(res_index, all_fits_iq[0, :] / 10 ** 6, 'o', label="IQ fit")

    bad_fit_mag = np.where(all_fits_mag[8, :] > reduced_chi_squared_cutoff)
    failed_fit_mag = np.where(all_fits_mag[0, :] == 0)
    bad_fit_mag = np.append(bad_fit_mag, failed_fit_mag)
    bad_fit_iq = np.where(all_fits_iq[9, :] > reduced_chi_squared_cutoff)
    failed_fit_iq = np.where(all_fits_iq[0, :] == 0)
    bad_fit_iq = np.append(bad_fit_mag, failed_fit_iq)

    plt.plot(res_index[bad_fit_mag], all_fits_mag[0, :][bad_fit_mag] / 10 ** 6, 'o', label="bad fit", color='k')
    plt.plot(res_index[bad_fit_iq], all_fits_iq[0, :][bad_fit_iq] / 10 ** 6, 'o', color='k')

    plt.xlabel("resonator index")
    plt.ylabel("Resonator Frequency (MHz)")
    plt.legend(loc=4)
    pdf_pages.savefig(fig)
    plt.close()

    fig = plt.figure(2, figsize=(12, 6))
    plt.title("Resonator Qs")
    plt.plot(res_index, all_fits_mag[1, :], 'o', label="Qr Mag", color='g')
    plt.plot(res_index, all_fits_iq[1, :], '*', label="Qr IQ", color='g')
    plt.plot(res_index[bad_fit_mag], all_fits_mag[1, :][bad_fit_mag], 'o', color='k')
    plt.plot(res_index[bad_fit_iq], all_fits_iq[1, :][bad_fit_iq], '*', color='k')

    plt.plot(res_index, all_fits_mag[1, :] / all_fits_mag[2, :], 'o', label="Qc Mag", color='b')
    plt.plot(res_index, all_fits_iq[1, :] / all_fits_mag[2, :], '*', label="Qc IQ", color='b')
    plt.plot(res_index[bad_fit_mag], all_fits_mag[1, :][bad_fit_mag] / all_fits_mag[2, :][bad_fit_mag], 'o', color='k')
    plt.plot(res_index[bad_fit_iq], all_fits_iq[1, :][bad_fit_iq] / all_fits_mag[2, :][bad_fit_iq], '*', color='k')

    plt.plot(res_index, 1 / (1 / all_fits_mag[1, :] - 1 / (all_fits_mag[1, :] / all_fits_mag[2, :])), 'o',
             label="Qi Mag", color='r')
    plt.plot(res_index, 1 / (1 / all_fits_iq[1, :] - 1 / (all_fits_iq[1, :] / all_fits_iq[2, :])), '*', label="Qi IQ",
             color='r')
    plt.plot(res_index[bad_fit_mag], 1 / (1 / all_fits_mag[1, :][bad_fit_mag] - 1 / (
                all_fits_mag[1, :][bad_fit_mag] / all_fits_mag[2, :][bad_fit_mag])), 'o', color='k')
    plt.plot(res_index[bad_fit_iq], 1 / (1 / all_fits_iq[1, :][bad_fit_iq] - 1 / (
                all_fits_iq[1, :][bad_fit_iq] / all_fits_iq[2, :][bad_fit_iq])), '*', color='k')

    plt.xlabel("Resonator index")
    plt.ylabel("Resonator Q")
    plt.yscale('log')
    plt.legend()
    pdf_pages.savefig(fig)
    plt.close()

    fig = plt.figure(3, figsize=(12, 6))
    plt.title("Non linearity parameter a")
    plt.plot(res_index, all_fits_mag[4, :], 'o', label="a Mag")
    plt.plot(res_index, all_fits_iq[4, :], 'o', label="a IQ")
    plt.plot(res_index[bad_fit_mag], all_fits_mag[4, :][bad_fit_mag], 'o', color='k')
    plt.plot(res_index[bad_fit_iq], all_fits_iq[4, :][bad_fit_iq], 'o', color='k')

    plt.xlabel("Resonator index")
    plt.ylabel("Non-linearity parameter a")
    plt.ylim(0, 1)
    plt.legend()
    pdf_pages.savefig(fig)
    plt.close()

    pdf_pages.close()

    np.save(outfile_dir + "/" + "all_fits_mag", all_fits_mag)
    np.save(outfile_dir + "/" + "all_fits_iq", all_fits_iq)
    np.savetxt(outfile_dir + "/" + "all_fits_mag.csv", all_fits_mag, delimiter=',')
    np.savetxt(outfile_dir + "/" + "all_fits_iq.csv", all_fits_iq, delimiter=',')


# this function fits a fine and gain scan combo produced by the ASU multitone system
def fit_fine_gain(fine_name, gain_name, fine_z_err=None, gain_z_err=None):
    fine = read_multitone.read_iq_sweep(fine_name)
    gain = read_multitone.read_iq_sweep(gain_name)
    outfile_dir = fine_name
    center_freqs = fine['freqs'][fine['freqs'].shape[0] // 2, :]

    pdf_pages = PdfPages(outfile_dir + "/" + "fit_plots.pdf")

    all_fits_mag = np.zeros((8, fine['freqs'].shape[1]))
    all_fits_iq = np.zeros((9, fine['freqs'].shape[1]))

    for i in range(0, fine['freqs'].shape[1]):
        fine_f = fine['freqs'][:, i] * 10 ** 6
        gain_f = gain['freqs'][:, i] * 10 ** 6
        fine_z = fine['I'][:, i] + 1.j * fine['Q'][:, i]
        gain_z = gain['I'][:, i] + 1.j * gain['Q'][:, i]

        # flag data that is too close to other resonators
        distance = center_freqs - center_freqs[i]
        if center_freqs[i] != np.min(center_freqs):  # don't do if lowest frequency resonator
            closest_lower_dist = -np.min(np.abs(distance[np.where(distance < 0)]))
            closest_lower_index = np.where(distance == closest_lower_dist)[0][0]
            halfway_low = (center_freqs[i] + center_freqs[closest_lower_index]) / 2.
        else:
            halfway_low = 0

        if center_freqs[i] != np.max(center_freqs):  # don't do if highest frequenct
            closest_higher_dist = np.min(np.abs(distance[np.where(distance > 0)]))
            closest_higher_index = np.where(distance == closest_higher_dist)[0][0]
            halfway_high = (center_freqs[i] + center_freqs[closest_higher_index]) / 2.
        else:
            halfway_high = np.inf

        use_index = np.where(((fine_f / 10 ** 6 > halfway_low) & (fine_f / 10 ** 6 < halfway_high)))
        fine_f = fine_f[use_index]
        fine_z = fine_z[use_index]

        # flag gain data that is to close to all of the resonators
        fine_span = (fine_f[-1] - fine_f[0]) / 10 ** 6
        use_boolean = np.zeros(len(gain_f))
        for j in range(0, len(gain_f)):
            if (np.abs(gain_f[j] / 10 ** 6 - center_freqs) < fine_span / 2.).any():
                use_boolean[j] = 1

        use_index_gain = np.where(use_boolean != 1)
        if len(use_index_gain[0] > 1):
            gain_f = gain_f[use_index_gain]
            gain_z = gain_z[use_index_gain]

        fig = plt.figure(i, figsize=(16, 10))

        plt.subplot(231)
        plt.title("Resonator Index " + str(i))
        plt.plot(fine['freqs'][:, i], 10 * np.log10(fine['I'][:, i] ** 2 + fine['Q'][:, i] ** 2), 'o', label="fine")
        plt.plot(gain['freqs'][:, i], 10 * np.log10(gain['I'][:, i] ** 2 + gain['Q'][:, i] ** 2), 'o', label="gain")
        plt.xlabel("Frequency (MHz)")
        plt.ylabel("Power (dB)")

        plt.subplot(232)
        plt.plot(fine['freqs'][:, i], 10 * np.log10(fine['I'][:, i] ** 2 + fine['Q'][:, i] ** 2), 'o')
        plt.plot(gain['freqs'][:, i], 10 * np.log10(gain['I'][:, i] ** 2 + gain['Q'][:, i] ** 2), 'o')
        plt.xlabel("Frequency (MHz)")
        plt.ylabel("Power (dB)")
        plt.xlim(np.min(fine['freqs'][:, i]), np.max(fine['freqs'][:, i]))

        # fit nonlinear magnitude
        try:
            x0 = guess_x0_mag_nonlinear_sep(fine_f, fine_z, gain_f, gain_z, verbose=True)
            fit_dict_mag = fit_nonlinear_mag_sep(fine_f, fine_z, gain_f, gain_z, fine_z_err=fine_z_err,
                                                 gain_z_err=gain_z_err, x0=x0)
            # ,bounds =bounds)
            all_fits_mag[:, i] = fit_dict_mag['fit'][0]
            plt.subplot(231)
            plt.plot(fit_dict_mag['fit_freqs'] / 10 ** 6, 10 * np.log10(fit_dict_mag['fit_result']), "+", label="fit")
            plt.plot(fit_dict_mag['fit_freqs'] / 10 ** 6, 10 * np.log10(fit_dict_mag['x0_result']), "x",
                     label="x0 guess")
            plt.title("f =" + str(fit_dict_mag['fit'][0][0] / 10 ** 6)[0:7] + "MHz, a=" + "{:.2f}".format(
                fit_dict_mag['fit'][0][4]))
            plt.legend()
            plt.subplot(232)
            plt.plot(fit_dict_mag['fit_freqs'] / 10 ** 6, 10 * np.log10(fit_dict_mag['fit_result']), "+")
            plt.plot(fit_dict_mag['fit_freqs'] / 10 ** 6, 10 * np.log10(fit_dict_mag['x0_result']), "x")

            plt.text(0.75, 0.85, "fr  = " + "{:.3f}".format(fit_dict_mag['fit'][0][0] / 10 ** 6) + " MHz", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.80, "Qr  = " + "{:.0f}".format(fit_dict_mag['fit'][0][1]) + " ", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.75, "amp = " + "{:.2f}".format(fit_dict_mag['fit'][0][2]) + " ", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.70, "phi = " + "{:.2f}".format(fit_dict_mag['fit'][0][3]) + " radians", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.65, "a   = " + "{:.2f}".format(fit_dict_mag['fit'][0][4]) + " ", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.60, "b0  = " + "{:.0f}".format(fit_dict_mag['fit'][0][5]) + " ", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.55, "b1  = " + "{:.0f}".format(fit_dict_mag['fit'][0][6]) + " ", fontsize=14,
                     transform=plt.gcf().transFigure)

        except Exception as e:
            print(e)
            print("could not fit the resonator")

        plt.subplot(234, aspect='equal')
        plt.plot(fine['I'][:, i], fine['Q'][:, i], 'o')
        plt.plot(gain['I'][:, i], gain['Q'][:, i], 'o')
        plt.xlabel("I")
        plt.ylabel("Q")

        plt.subplot(235, aspect='equal')
        plt.plot(fine['I'][:, i], fine['Q'][:, i], 'o')
        plt.plot(gain['I'][:, i], gain['Q'][:, i], 'o')
        plt.xlabel("I")
        plt.ylabel("Q")
        plt.xlim(np.min(fine['I'][:, i]), np.max(fine['I'][:, i]))
        plt.ylim(np.min(fine['Q'][:, i]), np.max(fine['Q'][:, i]))

        # fit nonlinear iq
        try:
            x0 = guess_x0_iq_nonlinear_sep(fine_f, fine_z, gain_f, gain_z, verbose=True)
            fit_dict_iq = fit_nonlinear_iq_sep(fine_f, fine_z, gain_f, gain_z, fine_z_err=fine_z_err,
                                               gain_z_err=gain_z_err, x0=x0)
            all_fits_iq[:, i] = fit_dict_iq['fit'][0]
            plt.subplot(234, aspect='equal')
            plt.plot(np.real(fit_dict_iq['fit_result']), np.imag(fit_dict_iq['fit_result']), "+")
            plt.plot(np.real(fit_dict_iq['x0_result']), np.imag(fit_dict_iq['x0_result']), "x")
            plt.title("f =" + str(fit_dict_iq['fit'][0][0] / 10 ** 6)[0:7] + "MHz, a=" + "{:.2f}".format(
                fit_dict_iq['fit'][0][4]))
            plt.subplot(235, aspect='equal')
            plt.plot(np.real(fit_dict_iq['fit_result']), np.imag(fit_dict_iq['fit_result']), "+")
            plt.plot(np.real(fit_dict_iq['x0_result']), np.imag(fit_dict_iq['x0_result']), "x")
            plt.plot(fine['I'][:, i], fine['Q'][:, i])

            plt.text(0.75, 0.45, "fr  = " + "{:.3f}".format(fit_dict_iq['fit'][0][0] / 10 ** 6) + " MHz", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.4, "Qr  = " + "{:.0f}".format(fit_dict_iq['fit'][0][1]) + " ", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.35, "amp = " + "{:.2f}".format(fit_dict_iq['fit'][0][2]) + " ", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.3, "phi = " + "{:.2f}".format(fit_dict_iq['fit'][0][3]) + " radians", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.25, "a   = " + "{:.2f}".format(fit_dict_iq['fit'][0][4]) + " ", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.2, "i0  = " + "{:.0f}".format(fit_dict_iq['fit'][0][5]) + " ", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.15, "q0  = " + "{:.0f}".format(fit_dict_iq['fit'][0][6]) + " ", fontsize=14,
                     transform=plt.gcf().transFigure)
            plt.text(0.75, 0.1, "tau = " + "{:.2f}".format(fit_dict_iq['fit'][0][7] * 10 ** 7) + " x 10^-7 ",
                     fontsize=14, transform=plt.gcf().transFigure)

        except Exception as e:
            print(e)
            print("could not fit the resonator")

        plt.suptitle("Resonator index = " + str(i) + ", Frequency = " + str(center_freqs[i])[0:7])

        pdf_pages.savefig(fig)
        plt.close(fig)

    pdf_pages.close()

    pdf_pages = PdfPages(outfile_dir + "/" + "fit_results.pdf")

    fig = plt.figure(1, figsize=(12, 6))
    plt.title("Center frequency")
    plt.plot(all_fits_mag[0, :] / 10 ** 6, 'o', label="Mag fit")
    plt.plot(all_fits_iq[0, :] / 10 ** 6, 'o', label="IQ fit")
    plt.xlabel("resonator index")
    plt.ylabel("Resonator Frequency (MHz)")
    plt.legend(loc=4)
    pdf_pages.savefig(fig)
    plt.close()

    fig = plt.figure(2, figsize=(12, 6))
    plt.title("Resonator Qs")
    plt.plot(all_fits_mag[1, :], 'o', label="Qr Mag", color='g')
    plt.plot(all_fits_iq[1, :], '*', label="Qr IQ", color='g')
    plt.plot(all_fits_mag[1, :] / all_fits_mag[2, :], 'o', label="Qc Mag", color='b')
    plt.plot(all_fits_iq[1, :] / all_fits_mag[2, :], '*', label="Qc IQ", color='b')
    plt.plot(1 / (1 / all_fits_mag[1, :] - 1 / (all_fits_mag[1, :] / all_fits_mag[2, :])), 'o', label="Qi Mag",
             color='r')
    plt.plot(1 / (1 / all_fits_iq[1, :] - 1 / (all_fits_iq[1, :] / all_fits_iq[2, :])), '*', label="Qi IQ", color='r')
    plt.xlabel("Resonator index")
    plt.ylabel("Resonator Q")
    plt.yscale('log')
    plt.legend()
    pdf_pages.savefig(fig)
    plt.close()

    fig = plt.figure(3, figsize=(12, 6))
    plt.title("Non linearity parameter a")
    plt.plot(all_fits_mag[4, :], 'o', label="a Mag")
    plt.plot(all_fits_iq[4, :], 'o', label="a IQ")
    plt.xlabel("Resonator index")
    plt.ylabel("Non-linearity parameter a")
    plt.ylim(0, 1)
    plt.legend()
    pdf_pages.savefig(fig)
    plt.close()

    pdf_pages.close()

    np.save(outfile_dir + "/" + "all_fits_mag", all_fits_mag)
    np.save(outfile_dir + "/" + "all_fits_iq", all_fits_iq)
    np.savetxt(outfile_dir + "/" + "all_fits_mag.csv", all_fits_mag, delimiter=',')
    np.savetxt(outfile_dir + "/" + "all_fits_iq.csv", all_fits_iq, delimiter=',')


def calibrate_list(fine_filename, gain_filename, stream_list, skip_beginning=0, bin_num=1, sample_rate=488.28125):
    # this is for batch fitting stream data in multiple dir files, mostly
    # for the beam map separate
    fine_dict = read_multitone.read_iq_sweep(fine_filename)
    gain_dict = read_multitone.read_iq_sweep(gain_filename)
    stream_dicts = [read_multitone.read_stream(stream_filename)
                    for stream_filename in stream_list]

    gain_z = gain_dict['I'] + 1.j * gain_dict['Q']
    fine_z = fine_dict['I'] + 1.j * fine_dict['Q']
    stream_z = [stream_dict['I_stream'][skip_beginning:] + 1.j * stream_dict['Q_stream'][skip_beginning:] for
                stream_dict in stream_dicts]
    stream_time = [np.asarray(stream_dict['packet_count'])[skip_beginning:] * 1 / sample_rate
                   for stream_dict in stream_dicts]
    stream_time = [time - time[0] for time in stream_time]

    # bin the data if you like
    if bin_num != 1:
        for i in range(0, stream_z.shape[1]):
            if i == 0:
                stream_z_downsamp = np.zeros((stream_z.shape[0] / bin_num, stream_z.shape[1]), dtype='complex')
            stream_z_downsamp[:, i] = np.mean(stream_z[0:stream_z.shape[0] / bin_num * bin_num, i].reshape(-1, bin_num),
                                              axis=1)  # int math
        stream_time_downsamp = np.mean(stream_time[0:stream_time.shape[0] / bin_num * bin_num].reshape(-1, bin_num),
                                       axis=1)  # int math
        stream_z = stream_z_downsamp
        stream_time = stream_time_downsamp

    # initalize some arrays to hold the calibrated data
    stream_corr_all = [np.zeros(set_z.shape, dtype='complex')
                       for set_z in stream_z]
    gain_corr_all = np.zeros(gain_z.shape, dtype='complex')
    fine_corr_all = np.zeros(fine_z.shape, dtype='complex')
    stream_df_over_f_all = [np.zeros(set_z.shape) for set_z in stream_z]
    fit_vals = np.zeros((3, fine_z.shape[1]))

    for k in range(0, min(fine_dict['I'].shape[1], stream_z[0].shape[1])):
        print(k)

        f_stream = fine_dict['freqs'][len(fine_dict['freqs'][:, k]) / 2, k] * 10 ** 6
        # normalize amplitude variation in gain scan
        amp_norm_dict = [amplitude_normalization_sep(
            gain_dict['freqs'][:, k] * 10 ** 6, gain_z[:, k],
            fine_dict['freqs'][:, k] * 10 ** 6, fine_z[:, k], f_stream,
            set_z[:, k]) for set_z in stream_z]

        # fit the gain
        gain_phase = np.arctan2(np.real(amp_norm_dict[0]['normalized_gain']),
                                np.imag(amp_norm_dict[0]['normalized_gain']))
        tau, fit_data_phase, gain_phase_rot = calibrate.fit_cable_delay(gain_dict['freqs'][:, k] * 10 ** 6, gain_phase)

        # remove cable delay
        gain_corr = calibrate.remove_cable_delay(gain_dict['freqs'][:, k] * 10 ** 6,
                                                 amp_norm_dict[0]['normalized_gain'], tau)
        fine_corr = calibrate.remove_cable_delay(fine_dict['freqs'][:, k] * 10 ** 6,
                                                 amp_norm_dict[0]['normalized_fine'], tau)
        stream_corr = [calibrate.remove_cable_delay(f_stream, set_norm_dict['normalized_stream'], tau) for set_norm_dict
                       in amp_norm_dict]

        # fit a cicle to the data
        xc, yc, R, residu = calibrate.leastsq_circle(np.real(fine_corr), np.imag(fine_corr))
        fit_vals[0, k] = xc
        fit_vals[1, k] = yc
        fit_vals[2, k] = R

        # move the data to the origin

        gain_corr = gain_corr - xc - 1j * yc
        fine_corr = fine_corr - xc - 1j * yc
        stream_corr = [set_corr - xc - 1j * yc for set_corr in stream_corr]

        # rotate so streaming data is at 0 pi
        phase_stream = [np.arctan2(np.imag(set_corr), np.real(set_corr))
                        for set_corr in stream_corr]
        med_phase = np.median(phase_stream)

        gain_corr_all[:, k] = gain_corr = gain_corr * np.exp(-1j * med_phase)
        fine_corr_all[:, k] = fine_corr = fine_corr * np.exp(-1j * med_phase)
        for i in range(len(stream_corr)):
            stream_corr_all[i][:, k] = stream_corr[i] = stream_corr[i] * np.exp(
                -1j * med_phase)

        phase_fine = np.arctan2(np.imag(fine_corr), np.real(fine_corr))
        use_index = np.where((-np.pi / 2. < phase_fine) & (phase_fine < np.pi / 2))
        phase_stream = [np.arctan2(np.imag(set_corr), np.real(set_corr))
                        for set_corr in stream_corr]

        # interp phase to frequency
        f_interp = interpolate.interp1d(phase_fine, fine_dict['freqs'][:, k], kind='quadratic', bounds_error=False,
                                        fill_value=0)

        phase_small = np.linspace(np.min(phase_fine), np.max(phase_fine), 1000)
        for i in range(len(phase_stream)):
            freqs_stream = f_interp(phase_stream[i])
            stream_df_over_f_all[i][:, k] = freqs_stream / np.mean(freqs_stream) - 1.

    # save everything to a dictionary
    cal_dict = {
        #                'fine_z': fine_z,
        #                'gain_z': gain_z,
        #                'stream_z': stream_z,
        #                'fine_freqs':fine_dict['freqs'],
        #                'gain_freqs':fine_dict['freqs'],
        #                'stream_corr':stream_corr_all,
        'gain_corr': gain_corr_all,
        'fine_corr': fine_corr_all,
        #                'stream_df_over_f':stream_df_over_f_all,
        #                'time':stream_dict['time'],
        #                'stream_time':stream_time}
        'fit_coords': fit_vals}

    # save the dictionary
    # pickle.dump( cal_dict, open(outfile_dir+ "cal.p", "wb" ),2 )
    return stream_df_over_f_all, stream_time, cal_dict


def calibrate_multi(fine_filename, gain_filename, stream_filename,
                    skip_beginning=0, plot_period=10, bin_num=1, outfile_dir="./",
                    sample_rate=488.28125, plot=True, **keywords):
    # read in the scans
    fine_dict = read_multitone.read_iq_sweep(fine_filename)
    gain_dict = read_multitone.read_iq_sweep(gain_filename)
    stream_dict = read_multitone.read_stream(stream_filename)

    # convert output to complex data
    gain_z = gain_dict['I'] + 1.j * gain_dict['Q']
    fine_z = fine_dict['I'] + 1.j * fine_dict['Q']
    stream_z = stream_dict['I_stream'][skip_beginning:] + 1.j * stream_dict['Q_stream'][skip_beginning:]
    # generate relative packet times
    stream_time = np.asarray(stream_dict['packet_count'])[skip_beginning:] * 1 / sample_rate
    stream_time = stream_time - stream_time[0]

    # bin the data if you like
    if bin_num != 1:
        stream_z_downsamp = sum([stream_z
                                 [:, i:bin_num * (stream_z.shape[1] // bin_num):bin_num]
                                 for i in range(bin_num)]) / bin_num
        stream_time_downsamp = sum([stream_time
                                    [:, i:bin_num * (stream_time.shape[1] // bin_num):bin_num]
                                    for i in range(bin_num)]) / bin_num
        stream_z = stream_z_downsamp
        stream_time = stream_time_downsamp

    # initalize some arrays to hold the calibrated data
    stream_corr_all = np.zeros(stream_z.shape, dtype='complex')
    gain_corr_all = np.zeros(gain_z.shape, dtype='complex')
    fine_corr_all = np.zeros(fine_z.shape, dtype='complex')
    stream_df_over_f_all = np.zeros(stream_z.shape)
    circle_fit = np.ndarray((fine_dict['I'].shape[1], 4))
    amp_dicts = []
    cable_delay_data = []

    for k in range(0, fine_dict['I'].shape[1]):
        print(k)

        f_stream = fine_dict['freqs'][len(fine_dict['freqs'][:, k]) / 2, k] * 10 ** 6
        # normalize amplitude varation in gain scan
        amp_norm_dict = amplitude_normalization_sep(
            gain_dict['freqs'][:, k] * 1e6,
            gain_z[:, k],
            fine_dict['freqs'][:, k] * 1e6,
            fine_z[:, k],
            f_stream,
            stream_z[:, k])
        amp_dicts.append(amp_norm_dict)

        # fit the gain
        gain_phase = np.arctan2(np.real(amp_norm_dict['normalized_gain']),
                                np.imag(amp_norm_dict['normalized_gain']))
        tau, fit_data_phase, gain_phase_rot = calibrate.fit_cable_delay(
            gain_dict['freqs'][:, k] * 1e6, gain_phase)
        cable_delay_data.append(
            (gain_phase, tau, fit_data_phase, gain_phase_rot))

        # remove cable delay
        gain_corr = calibrate.remove_cable_delay(gain_dict['freqs'][:, k] * 1e6, amp_norm_dict['normalized_gain'], tau)
        fine_corr = calibrate.remove_cable_delay(fine_dict['freqs'][:, k] * 1e6, amp_norm_dict['normalized_fine'], tau)
        stream_corr = calibrate.remove_cable_delay(f_stream, amp_norm_dict['normalized_stream'], tau)

        # fit a cicle to the data
        xc, yc, R, residu = calibrate.leastsq_circle(np.real(fine_corr), np.imag(fine_corr))
        circle_fit[k, 0:3] = np.array([xc, yc, R])

        # move the data to the origin
        gain_corr = gain_corr - xc - 1j * yc
        fine_corr = fine_corr - xc - 1j * yc
        stream_corr = stream_corr - xc - 1j * yc

        # rotate so streaming data is at 0 pi
        phase_stream = np.arctan2(np.imag(stream_corr), np.real(stream_corr))
        if ("rotate_fine_first" in keywords):  # if you have data that covers a large part of the iq loop
            med_phase = np.arctan2(np.imag(fine_corr), np.real(fine_corr))[0] + np.pi
        else:
            med_phase = np.median(phase_stream)
        circle_fit[k, -1] = med_phase

        gain_corr_all[:, k] = gain_corr = gain_corr * np.exp(-1j * med_phase)
        fine_corr_all[:, k] = fine_corr = fine_corr * np.exp(-1j * med_phase)
        stream_corr_all[:, k] = stream_corr = stream_corr * np.exp(-1j * med_phase)

        phase_fine = np.arctan2(np.imag(fine_corr), np.real(fine_corr))
        use_index = np.where((-np.pi / 2. < phase_fine) & (phase_fine < np.pi / 2))
        phase_stream = np.arctan2(np.imag(stream_corr), np.real(stream_corr))

        # interp phase to frequency
        f_interp = interpolate.interp1d(phase_fine, fine_dict['freqs'][:, k], kind='quadratic', bounds_error=False,
                                        fill_value=0)

        phase_small = np.linspace(np.min(phase_fine), np.max(phase_fine), 1000)
        freqs_stream = f_interp(phase_stream)
        stream_df_over_f_all[:, k] = stream_df_over_f = freqs_stream / np.mean(freqs_stream) - 1.

    # save everything to a dictionary
    cal_dict = {'fine_z': fine_z,
                'gain_z': gain_z,
                'stream_z': stream_z,
                'fine_freqs': fine_dict['freqs'],
                'gain_freqs': gain_dict['freqs'],
                'stream_corr': stream_corr_all,
                'gain_corr': gain_corr_all,
                'fine_corr': fine_corr_all,
                'stream_df_over_f': stream_df_over_f_all,
                'time': stream_dict['time'],
                'stream_time': stream_time}

    # plot the data if desired
    if plot:
        plot_calibrate(cal_dict, circle_fit, amp_dicts, cable_delay_data,
                       plot_period, outfile_dir)

    # save the dictionary
    pickle.dump(cal_dict, open(outfile_dir + "cal.p", "wb"), 2)
    return cal_dict


def plot_calibrate(cal_dict, circle_fit, amp_dicts, cable_delay_data,
                   plot_period, outfile_dir='./'):
    pdf_pages = PdfPages(outfile_dir + "cal_plots.pdf")
    for k in range(cal_dict['fine_z'].shape[1]):
        fig = plt.figure(k, figsize=(16, 10))

        # plot the raw data
        plt.subplot(241, aspect='equal')
        plt.title("Raw data")
        plt.plot(np.real(cal_dict['fine_z'][:, k]),
                 np.imag(cal_dict['fine_z'][:, k]), 'o')
        plt.plot(np.real(cal_dict['stream_z'][:, k][::plot_period]),
                 np.imag(cal_dict['stream_z'][:, k][::plot_period]), '.')
        plt.plot(np.real(cal_dict['gain_z'])[:, k],
                 np.imag(cal_dict['gain_z'])[:, k], 'o')

        # plot fine and stream data
        plt.subplot(242)
        plt.title("Raw data")
        plt.plot(np.real(cal_dict['fine_z'])[:, k],
                 np.imag(cal_dict['fine_z'])[:, k], 'o')
        plt.plot(np.real(cal_dict['stream_z'][:, k][::plot_period]),
                 np.imag(cal_dict['stream_z'][:, k][::plot_period]), '.')

        # plot the fit of the gain amp/phase variation
        plt.subplot(243)
        plt.title("Gain amplitude variation fit")
        plt.plot(cal_dict['gain_freqs'][:, k] * 1e6,
                 20. * np.log10(np.abs(cal_dict['gain_z'][:, k])), 'o')
        plt.plot(cal_dict['gain_freqs'][:, k] * 1e6,
                 10. * np.log10(np.abs(amp_dicts[k]['normalized_gain']) ** 2), 'o')
        plt.plot(cal_dict['fine_freqs'][:, k] * 1e6,
                 10. * np.log10(np.abs(amp_dicts[k]['normalized_fine']) ** 2), 'o')
        plt.plot(cal_dict['gain_freqs'][:, k] * 1e6,
                 20. * np.log10(np.abs(amp_dicts[k]['poly_data'])))

        # plot gain amp variation
        plt.subplot(244)
        plt.title("Data nomalized for gain amplitude variation")
        plt.plot(np.real(amp_dicts[k]['normalized_fine']),
                 np.imag(amp_dicts[k]['normalized_fine']), 'o')
        plt.plot(np.real(amp_dicts[k]['normalized_stream'][::plot_period]),
                 np.imag(amp_dicts[k]['normalized_stream'][::plot_period]), '.')

        gain_phase, tau, fit_data_phase, gain_phase_rot = cable_delay_data[k]
        # plot gain phase fit
        plt.subplot(245)
        plt.title("Gain phase fit")
        plt.plot(cal_dict['gain_freqs'][:, k], gain_phase_rot, 'o')
        plt.plot(cal_dict['gain_freqs'][:, k], fit_data_phase)
        plt.xlabel("Frequency (MHz)")
        plt.ylabel("Phase")

        # unpack corr data
        gain_corr = cal_dict['gain_corr'][:, k]
        fine_corr = cal_dict['fine_corr'][:, k]
        stream_corr = cal_dict['stream_corr'][:, k]
        # reverse the circle offset and rotation
        med_phase = circle_fit[k, -1]
        xc = circle_fit[k, 0]
        yc = circle_fit[k, 1]
        R = circle_fit[k, 2]
        gain_corr0 = gain_corr * np.exp(1.j * med_phase) + xc + 1.j * yc
        fine_corr0 = fine_corr * np.exp(1.j * med_phase) + xc + 1.j * yc
        stream_corr0 = stream_corr * np.exp(1.j * med_phase) + xc + 1.j * yc

        plt.subplot(246)
        plt.title("Cable delay removed")
        plt.plot(np.real(gain_corr0), np.imag(gain_corr0), 'o')
        plt.plot(np.real(fine_corr0), np.imag(fine_corr0), 'o')
        plt.plot(np.real(stream_corr0)[10:-10][::plot_period],
                 np.imag(stream_corr0)[10:-10][::plot_period], '.')

        # center and rotate IQ circle
        plt.subplot(247)
        plt.title("Moved to 0,0 and rotated")
        plt.plot(np.real(stream_corr)[2:-1][::plot_period],
                 np.imag(stream_corr)[2:-1][::plot_period], '.')
        plt.plot(np.real(gain_corr), np.imag(gain_corr), 'o')
        plt.plot(np.real(fine_corr), np.imag(fine_corr), 'o')
        calibrate.plot_data_circle(np.real(fine_corr) - xc, np.imag(fine_corr) - yc,
                                   0, 0, R)

        # redo the phase fitting
        phase_fine = np.arctan2(np.imag(fine_corr), np.real(fine_corr))
        use_index = np.where((-np.pi / 2. < phase_fine) & (phase_fine < np.pi / 2))
        phase_stream = np.arctan2(np.imag(stream_corr), np.real(stream_corr))

        # interp phase to frequency
        f_interp = interpolate.interp1d(phase_fine, cal_dict['fine_freqs'][:, k],
                                        kind='quadratic', bounds_error=False, fill_value=0)

        phase_small = np.linspace(np.min(phase_fine), np.max(phase_fine), 1000)
        freqs_stream = f_interp(phase_stream)

        # plot the stream phase
        plt.subplot(248)
        plt.plot(phase_fine, cal_dict['fine_freqs'][:, k], 'o')
        plt.plot(phase_small, f_interp(phase_small), '--')
        plt.plot(phase_stream[::plot_period], freqs_stream[::plot_period], '.')
        plt.ylim(np.min(freqs_stream) - (np.max(freqs_stream) - np.min(freqs_stream)) * 3,
                 np.max(freqs_stream) + (np.max(freqs_stream) - np.min(freqs_stream)) * 3)
        plt.xlim(np.min(phase_stream) - np.pi / 4, np.max(phase_stream) + np.pi / 4)
        plt.xlabel("phase")
        plt.ylabel("Frequency")

        pdf_pages.savefig(fig)
        plt.close(fig)

    pdf_pages.close()


def fft_noise(z_stream, df_over_f, sample_rate):
    npts_fft = int(2 ** (np.floor(np.log2(df_over_f.size))))
    Sxx = 2 * fftpack.fft(df_over_f, n=npts_fft) * np.conj(
        fftpack.fft(df_over_f, n=npts_fft)) / sample_rate * npts_fft / npts_fft ** 2
    # perpendicular should be radial on the circle
    per_stream = np.abs(z_stream)
    # radius times angle should be distance along the circle and should work if
    # noise ball is not small
    par_stream = np.mean(per_stream) * np.arctan2(np.imag(z_stream), np.real(z_stream))
    S_per = 2 * fftpack.fft(per_stream, n=npts_fft) * np.conj(
        fftpack.fft(per_stream, n=npts_fft)
    ) / sample_rate * npts_fft / npts_fft ** 2
    S_par = 2 * fftpack.fft(par_stream, n=npts_fft) * np.conj(
        fftpack.fft(par_stream, n=npts_fft)) / sample_rate * npts_fft / npts_fft ** 2
    fft_freqs = fftpack.fftfreq(npts_fft, 1. / sample_rate)
    return fft_freqs, Sxx, S_per, S_par


def noise_multi(cal_dict, sample_rate=488.28125, outfile_dir="./", n_comp_PCA=0):
    if n_comp_PCA > 0:
        do_PCA = True
        # do PCA on the data
        cleaned, removed = PCA.PCA_SVD(cal_dict['stream_df_over_f'], n_comp_PCA,
                                       plot=True)
    else:
        do_PCA = False

    for k in range(0, cal_dict['fine_corr'].shape[1]):
        print(k)

        # lets fourier transfer that crap
        fft_freqs, Sxx, S_per, S_par = fft_noise(cal_dict['stream_corr'][:, k], cal_dict['stream_df_over_f'][:, k],
                                                 sample_rate)
        if do_PCA:
            fft_freqs_2, Sxx_clean, S_per_2, S_par_2 = fft_noise(cal_dict['stream_corr'][:, k], cleaned[:, k],
                                                                 sample_rate)
        if k == 0:
            # intialize some arrays
            Sxx_all = np.zeros((Sxx.shape[0], cal_dict['fine_corr'].shape[1]))
            Sxx_all_clean = np.zeros((Sxx.shape[0], cal_dict['fine_corr'].shape[1]))
            S_per_all = np.zeros((S_per.shape[0], cal_dict['fine_corr'].shape[1]))
            S_par_all = np.zeros((S_par.shape[0], cal_dict['fine_corr'].shape[1]))

        Sxx_all[:, k] = np.abs(Sxx)
        if do_PCA:
            Sxx_all_clean[:, k] = np.abs(Sxx_clean)
        S_per_all[:, k] = np.abs(S_per)
        S_par_all[:, k] = np.abs(S_par)

        # bin it for ploting
        plot_bins = np.logspace(-3, np.log10(250), 100)
        binnedfreq = binned_statistic(fft_freqs, fft_freqs, bins=plot_bins)[0]  # bin the frequecy against itself
        binnedpsd = binned_statistic(fft_freqs, np.abs(Sxx), bins=plot_bins)[0]
        if do_PCA:
            binnedpsd_clean = binned_statistic(fft_freqs, np.abs(Sxx_clean), bins=plot_bins)[0]
        binnedper = binned_statistic(fft_freqs, np.abs(S_per), bins=plot_bins)[0]
        binnedpar = binned_statistic(fft_freqs, np.abs(S_par), bins=plot_bins)[0]
        amp_subtracted = np.abs(binnedpsd) * (binnedpar - binnedper) / binnedpar
        if k == 0:
            Sxx_binned_all = np.zeros((binnedfreq.shape[0], cal_dict['fine_corr'].shape[1]))
            Sxx_binned_all_clean = np.zeros((binnedfreq.shape[0], cal_dict['fine_corr'].shape[1]))
            S_per_binned_all = np.zeros((binnedfreq.shape[0], cal_dict['fine_corr'].shape[1]))
            S_par_binned_all = np.zeros((binnedfreq.shape[0], cal_dict['fine_corr'].shape[1]))
            amp_subtracted_all = np.zeros((binnedfreq.shape[0], cal_dict['fine_corr'].shape[1]))

        Sxx_binned_all[:, k] = binnedpsd
        if do_PCA:
            Sxx_binned_all_clean[:, k] = binnedpsd_clean
        S_per_binned_all[:, k] = binnedper
        S_par_binned_all[:, k] = binnedpar
        amp_subtracted_all[:, k] = amp_subtracted

    # make a psd dictionary
    psd_dict = {'fft_freqs': fft_freqs,
                'Sxx': Sxx_all,
                'S_per': S_per_all,
                'S_par': S_par_all,
                'binned_freqs': binnedfreq,
                'Sxx_binned': Sxx_binned_all,
                'S_per_binned': S_per_binned_all,
                'S_par_binned': S_par_binned_all,
                'amp_subtracted': amp_subtracted_all,
                'Sxx_clean': Sxx_all_clean,
                'Sxx_binned_clean': Sxx_binned_all_clean}

    # plot the stuff
    plot_noise_multi(psd_dict, N_PCA=n_comp_PCA, outfile_dir=outfile_dir)
    # save the psd dictionary
    pickle.dump(psd_dict, open(outfile_dir + "psd.p", "wb"), 2)

    return psd_dict


def plot_noise_multi(psd_dict, white_avg=100., N_PCA=0, outfile_dir="./"):
    # create the PDF file
    pdf_pages = PdfPages(outfile_dir + "psd_plots.pdf")
    # was there a PCA analysis
    do_PCA = (N_PCA > 0)
    # calculate white noise levels and plot them
    freq_mask = psd_dict['fft_freqs'] > white_avg
    N_res = psd_dict['Sxx'].shape[1]
    Sxx_avg = np.ndarray((N_res,))
    par_per_ratio = np.ndarray((N_res,))
    for i in range(N_res):
        Sxx_avg[i] = np.mean(psd_dict['Sxx'][:, i][freq_mask])
        par_per_ratio[i] = np.mean(psd_dict['S_par'][:, i][freq_mask]
                                   ) / np.mean(psd_dict['S_per'][:, i][freq_mask])
    fig = plt.figure(9000, figsize=(16, 6))
    plt.subplot(122)
    plt.title("White Noise Levels")
    plt.semilogy(np.arange(N_res), Sxx_avg, 'bo')
    plt.ylabel("Sxx (1/Hz)")
    plt.xlabel("Resonator Index")
    plt.subplot(121)
    plt.title("Total to Amplifier Ratio")
    plt.plot(np.arange(N_res), par_per_ratio, 'go')
    plt.ylabel("Ratio of Parallel to Perpendicular")
    plt.xlabel("Resonator Index")
    pdf_pages.savefig(fig)
    plt.close(fig)
    # loop over the resonators
    for k in range(N_res):
        fig = plt.figure(k, figsize=(16, 6))
        # plot Sxx raw and with amplifier subtracted
        plt.subplot(122)
        plt.title("Sxx")
        plt.loglog(psd_dict['binned_freqs'],
                   np.abs(psd_dict['Sxx_binned'][:, k]), linewidth=2,
                   label="Sxx raw")
        if do_PCA:
            plt.loglog(psd_dict['binned_freqs'],
                       np.abs(psd_dict['Sxx_binned_clean'][:, k]), linewidth=2,
                       label="PCA {0} comps".format(N_PCA))
        plt.loglog(psd_dict['binned_freqs'],
                   psd_dict['amp_subtracted'][:, k],
                   linewidth=2, label="raw amp subtracted")
        plt.ylabel("Sxx (1/Hz)")
        plt.xlabel("Frequency (Hz)")
        plt.legend()

        # plot noise independant quadratures
        plt.subplot(121)
        plt.title("Res indx = " + str(k))

        plt.loglog(psd_dict['binned_freqs'], psd_dict['S_per_binned'][:, k],
                   label="amp noise")
        plt.loglog(psd_dict['binned_freqs'], psd_dict['S_par_binned'][:, k],
                   label="detect noise")
        plt.legend()
        plt.xlabel("Frequency (Hz)")

        # save and close the figure
        pdf_pages.savefig(fig)
        plt.close(fig)

    # close the pdf file so everything gets flushed to the disk
    pdf_pages.close()



# Last Updated: 02/07/2018

import visa


class Anritsu(object):

    # Default constructor used to call the instrument at the given
    # GPIBaddress and initializes the object to the instrument
    def __init__(self):
        self.GPIBaddress = 'GPIB0::10::INSTR'
        rm = visa.ResourceManager("@py")
        self.obj1 = rm.open_resource(self.GPIBaddress)

    # This method returns the CURRENT frequency that Anritsu is at
    # Using the built in command - OF1
    def get_frequency(self):
        frequency = float(self.obj1.query('OF1'))
        return frequency * 10 ** 6

    # This method will set the frequency of the Anritsu to the
    # specified amount. The given freq_value needs to be within
    # the range of 10MHz - 20GHz.
    def set_frequency(self, freq_value):

        # If value is below/above a certain threshold,
        # let user know to re-enter

        if (freq_value < 10000000):
            print(freq_value)
            print("Invalid value - Needs to be greater than 10MHz")
        elif (freq_value > 20000000000):
            print("Invalid value - needs to be smaller than 20GHz")
        else:
            cmd_str = 'CF1 %9.9fGH' % (freq_value / 1e9)
            self.obj1.write(cmd_str)
        freq = self.get_frequency()
        print(freq)

    # This method will return the CURRENT power that the frequency
    # is on
    def get_power(self):
        power = float(self.obj1.query('OL1'))
        return power

    # This method will set the power to the specified amount
    def set_power(self, power_value):

        # If the value is below/above a certain threshold,
        # let user know to re-enter
        if (power_value < -20):
            print("Invalid value - Needs to be greater than -20dbm")
        elif (power_value > 30):
            print("Invalid value - Needs to be less than 30dbm")
        else:
            cmd_str = 'L1 %3.2fDM' % power_value
            self.obj1.write(cmd_str)

    # This method will turn the output to ON
    def turn_outputOn(self):
        self.obj1.write('RF1')

    # This method will turn the output to OFF
    def turn_outputOff(self):
        self.obj1.write('RF0')

    # This method will print out to terminal all of the available
    # functions that are included in this class
    def help(self):
        print("")
        print("\t Available Functions")
        print("\t --------------------")
        print("\t set_frequency(Hz)")
        print("\t get_frequency()")
        print("\t get_power()")
        print("\t set_power(Dbm)")
        print("\t turn_outputOn")
        print("\t turn_outputOff")
